# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>
# Copyright (C) Airbus DS CyberSecurity
# Authors: Jean-Michel Picod, Arnaud Lebrun, Jonathan Christofer Demay

"""Bluetooth 4LE layer"""

import struct

from scapy.compat import orb, chb
from scapy.config import conf
from scapy.data import (
    DLT_BLUETOOTH_LE_LL,
    DLT_BLUETOOTH_LE_LL_WITH_PHDR,
    PPI_BTLE,
)
from scapy.packet import Packet, bind_layers
from scapy.fields import (
    BitEnumField,
    BitField,
    ByteEnumField,
    ByteField,
    Field,
    FlagsField,
    LEIntField,
    LEShortEnumField,
    LEShortField,
    MACField,
    PacketListField,
    SignedByteField,
    X3BytesField,
    XByteField,
    XIntField,
    XLEIntField,
    XLELongField,
    XLEShortField,
    XShortField,
)
from scapy.contrib.ethercat import LEBitEnumField, LEBitField

from scapy.layers.bluetooth import EIR_Hdr, L2CAP_Hdr
from scapy.layers.ppi import PPI_Element, PPI_Hdr

from scapy.utils import mac2str, str2mac

####################
# Transport Layers #
####################


class BTLE_PPI(PPI_Element):
    """Cooked BTLE PPI header

    See ``ppi_btle_t`` in
    https://github.com/greatscottgadgets/libbtbb/blob/master/lib/src/pcap.c
    """
    name = "BTLE PPI header"
    fields_desc = [
        ByteField("btle_version", 0),
        # btle_channel is a frequency in MHz. Named for consistency with
        # other users.
        LEShortField("btle_channel", None),
        ByteField("btle_clkn_high", None),
        LEIntField("btle_clk_100ns", None),
        SignedByteField("rssi_max", None),
        SignedByteField("rssi_min", None),
        SignedByteField("rssi_avg", None),
        ByteField("rssi_count", None)
    ]


class BTLE_RF(Packet):
    """Cooked BTLE link-layer pseudoheader.

    https://www.tcpdump.org/linktypes/LINKTYPE_BLUETOOTH_LE_LL_WITH_PHDR.html
    """
    name = "BTLE RF info header"

    _TYPES = {
        0: "ADV_OR_DATA_UNKNOWN_DIR",
        1: "AUX_ADV",
        2: "DATA_M_TO_S",
        3: "DATA_S_TO_M",
        4: "CONN_ISO_M_TO_S",
        5: "CONN_ISO_S_TO_M",
        6: "BROADCAST_ISO",
        7: "RFU",
    }

    _PHY = {
        0: "1M",
        1: "2M",
        2: "Coded",
        3: "RFU",
    }

    fields_desc = [
        ByteField("rf_channel", 0),
        SignedByteField("signal", -128),
        SignedByteField("noise", -128),
        ByteField("access_address_offenses", 0),
        XLEIntField("reference_access_address", 0),
        LEBitField("dewhitened", 0, 1),
        LEBitField("sig_power_valid", 0, 1),
        LEBitField("noise_power_valid", 0, 1),
        LEBitField("decrypted", 0, 1),
        LEBitField("reference_access_address_valid", 0, 1),
        LEBitField("access_address_offenses_valid", 0, 1),
        LEBitField("channel_aliased", 0, 1),
        LEBitEnumField("type", 0, 3, _TYPES),
        LEBitField("crc_checked", 0, 1),
        LEBitField("crc_valid", 0, 1),
        LEBitField("mic_checked", 0, 1),
        LEBitField("mic_valid", 0, 1),
        LEBitEnumField("phy", 0, 2, _PHY),
    ]


##########
# Fields #
##########

class BDAddrField(MACField):
    def __init__(self, name, default, resolve=False):
        MACField.__init__(self, name, default)
        if resolve:
            conf.resolve.add(self)

    def i2m(self, pkt, x):
        if x is None:
            return b"\0\0\0\0\0\0"
        return mac2str(':'.join(x.split(':')[::-1]))

    def m2i(self, pkt, x):
        return str2mac(x[::-1])


class BTLEChanMapField(XByteField):
    def __init__(self, name, default):
        Field.__init__(self, name, default, "<Q")

    def addfield(self, pkt, s, val):
        return s + struct.pack(self.fmt, self.i2m(pkt, val))[:5]

    def getfield(self, pkt, s):
        return s[5:], self.m2i(pkt, struct.unpack(self.fmt, s[:5] + b"\x00\x00\x00")[0])  # noqa: E501


class BTLEFeatureField(FlagsField):
    def __init__(self, name, default):
        super(BTLEFeatureField, self).__init__(
            name, default, -64,
            ['le_encryption',
             'conn_par_req_proc',
             'ext_reject_ind',
             'slave_init_feat_exch',
             'le_ping',
             'le_data_len_ext',
             'll_privacy',
             'ext_scan_filter',
             'le_2m_phy',
             'tx_mod_idx',
             'rx_mod_idx',
             'le_coded_phy',
             'le_ext_adv',
             'le_periodic_adv',
             'ch_sel_alg',
             'le_pwr_class'
             'min_used_channels',
             'conn_cte_req',
             'conn_cte_rsp',
             'connless_cte_tx',
             'connless_cte_rx',
             'antenna_switching_cte_aod_tx',
             'antenna_switching_cte_aoa_rx',
             'cte_rx',
             'periodic_adv_sync_transfer_tx',
             'periodic_adv_sync_transfer_rx',
             'sleep_clock_accuracy_updates',
             'remote_public_key_validation',
             'cis_central',
             'cis_peripheral',
             'iso_broadcaster',
             'synchronized_receiver',
             'connected_iso_host_support',
             'le_power_control_request',
             'le_power_control_request',
             'le_path_loss_monitoring',
             'periodic_adv_adi_support',
             'connection_subrating',
             'connection_subrating_host_support',
             'channel_classification']
        )


class BTLEPhysField(FlagsField):
    def __init__(self, name, default):
        super(BTLEPhysField, self).__init__(
            name, default, -8,
            ['phy_1m', 'phy_2m', 'phy_coded']
        )


##########
# Layers #
##########

class BTLE(Packet):
    name = "BT4LE"
    fields_desc = [
        XLEIntField("access_addr", 0x8E89BED6),
        X3BytesField("crc", None)
    ]

    @staticmethod
    def compute_crc(pdu, init=0x555555):
        def swapbits(a):
            v = 0
            if a & 0x80 != 0:
                v |= 0x01
            if a & 0x40 != 0:
                v |= 0x02
            if a & 0x20 != 0:
                v |= 0x04
            if a & 0x10 != 0:
                v |= 0x08
            if a & 0x08 != 0:
                v |= 0x10
            if a & 0x04 != 0:
                v |= 0x20
            if a & 0x02 != 0:
                v |= 0x40
            if a & 0x01 != 0:
                v |= 0x80
            return v

        state = swapbits(init & 0xff) + (swapbits((init >> 8) & 0xff) << 8) + (swapbits((init >> 16) & 0xff) << 16)  # noqa: E501
        lfsr_mask = 0x5a6000
        for i in (orb(x) for x in pdu):
            for j in range(8):
                next_bit = (state ^ i) & 1
                i >>= 1
                state >>= 1
                if next_bit:
                    state |= 1 << 23
                    state ^= lfsr_mask
        return struct.pack("<L", state)[:-1]

    def post_build(self, p, pay):
        # Switch payload and CRC
        crc = p[-3:]
        p = p[:-3] + pay
        p += crc if self.crc is not None else self.compute_crc(p[4:])
        return p

    def post_dissect(self, s):
        self.raw_packet_cache = None  # Reset packet to allow post_build
        return s

    def pre_dissect(self, s):
        # move crc
        return s[:4] + s[-3:] + s[4:-3]

    def hashret(self):
        return struct.pack("!L", self.access_addr)


class BTLE_ADV(Packet):
    # BT Core 5.2 - 2.3 ADVERTISING PHYSICAL CHANNEL PDU
    name = "BTLE advertising header"
    fields_desc = [
        BitEnumField("RxAdd", 0, 1, {0: "public",
                                     1: "random"}),
        BitEnumField("TxAdd", 0, 1, {0: "public",
                                     1: "random"}),
        # 4.5.8.3.1 - LE Channel Selection Algorithm #2
        BitEnumField("ChSel", 0, 1, {1: "#2"}),
        BitField("RFU", 0, 1),  # Unused
        BitEnumField("PDU_type", 0, 4, {0: "ADV_IND",
                                        1: "ADV_DIRECT_IND",
                                        2: "ADV_NONCONN_IND",
                                        3: "SCAN_REQ",
                                        4: "SCAN_RSP",
                                        5: "CONNECT_IND",
                                        6: "ADV_SCAN_IND"}),
        XByteField("Length", None),
    ]

    def post_build(self, p, pay):
        p += pay
        if self.Length is None:
            if len(pay) > 2:
                l_pay = len(pay)
            else:
                l_pay = 0
            p = p[:1] + chb(l_pay & 0xff) + p[2:]
        if not isinstance(self.underlayer, BTLE):
            self.add_underlayer(BTLE)
        return p


class BTLE_DATA(Packet):
    name = "BTLE data header"
    fields_desc = [
        BitField("RFU", 0, 3),  # Unused
        BitField("MD", 0, 1),
        BitField("SN", 0, 1),
        BitField("NESN", 0, 1),
        BitEnumField("LLID", 0, 2, {1: "continue", 2: "start", 3: "control"}),
        ByteField("len", None),
    ]

    def post_build(self, p, pay):
        if self.len is None:
            p = p[:-1] + chb(len(pay))
        return p + pay


class BTLE_ADV_IND(Packet):
    name = "BTLE ADV_IND"
    fields_desc = [
        BDAddrField("AdvA", None),
        PacketListField("data", None, EIR_Hdr)
    ]


class BTLE_ADV_DIRECT_IND(Packet):
    name = "BTLE ADV_DIRECT_IND"
    fields_desc = [
        BDAddrField("AdvA", None),
        BDAddrField("InitA", None)
    ]


class BTLE_ADV_NONCONN_IND(BTLE_ADV_IND):
    name = "BTLE ADV_NONCONN_IND"


class BTLE_ADV_SCAN_IND(BTLE_ADV_IND):
    name = "BTLE ADV_SCAN_IND"


class BTLE_SCAN_REQ(Packet):
    name = "BTLE scan request"
    fields_desc = [
        BDAddrField("ScanA", None),
        BDAddrField("AdvA", None)
    ]

    def answers(self, other):
        return BTLE_SCAN_RSP in other and self.AdvA == other.AdvA


class BTLE_SCAN_RSP(Packet):
    name = "BTLE scan response"
    fields_desc = [
        BDAddrField("AdvA", None),
        PacketListField("data", None, EIR_Hdr)
    ]

    def answers(self, other):
        return BTLE_SCAN_REQ in other and self.AdvA == other.AdvA


class BTLE_CONNECT_REQ(Packet):
    name = "BTLE connect request"
    fields_desc = [
        BDAddrField("InitA", None),
        BDAddrField("AdvA", None),
        # LLDATA
        XIntField("AA", 0x00),
        X3BytesField("crc_init", 0x0),
        XByteField("win_size", 0x0),
        XLEShortField("win_offset", 0x0),
        XLEShortField("interval", 0x0),
        XLEShortField("latency", 0x0),
        XLEShortField("timeout", 0x0),
        BTLEChanMapField("chM", 0),
        BitField("SCA", 0, 3),
        BitField("hop", 0, 5),
    ]


BTLE_Versions = {
    6: '4.0',
    7: '4.1',
    8: '4.2',
    9: '5.0',
    10: '5.1',
    11: '5.2',
}


BTLE_Corp_IDs = {
    0xf: 'Broadcom Corporation',
    0x59: 'Nordic Semiconductor ASA'
}


BTLE_BTLE_CTRL_opcode = {
    0x00: 'LL_CONNECTION_UPDATE_REQ',
    0x01: 'LL_CHANNEL_MAP_REQ',
    0x02: 'LL_TERMINATE_IND',
    0x03: 'LL_ENC_REQ',
    0x04: 'LL_ENC_RSP',
    0x05: 'LL_START_ENC_REQ',
    0x06: 'LL_START_ENC_RSP',
    0x07: 'LL_UNKNOWN_RSP',
    0x08: 'LL_FEATURE_REQ',
    0x09: 'LL_FEATURE_RSP',
    0x0A: 'LL_PAUSE_ENC_REQ',
    0x0B: 'LL_PAUSE_ENC_RSP',
    0x0C: 'LL_VERSION_IND',
    0x0D: 'LL_REJECT_IND',
    0x0E: 'LL_SLAVE_FEATURE_REQ',
    0x0F: 'LL_CONNECTION_PARAM_REQ',
    0x10: 'LL_CONNECTION_PARAM_RSP',
    0x14: 'LL_LENGTH_REQ',
    0x15: 'LL_LENGTH_RSP',
    0x16: 'LL_PHY_REQ',
    0x17: 'LL_PHY_RSP',
    0x18: 'LL_PHY_UPDATE_IND',
    0x19: 'LL_MIN_USED_CHANNELS',
    0x1A: 'LL_CTE_REQ',
    0x1B: 'LL_CTE_RSP',
    0x1C: 'LL_PERIODIC_SYNC_IND',
    0x1D: 'LL_CLOCK_ACCURACY_REQ',
    0x1E: 'LL_CLOCK_ACCURACY_RSP',
    0x1F: 'LL_CIS_REQ',
    0x20: 'LL_CIS_RSP',
    0x21: 'LL_CIS_IND',
    0x22: 'LL_CIS_TERMINATE_IND',
    0x23: 'LL_POWER_CONTROL_REQ',
    0x24: 'LL_POWER_CONTROL_RSP',
    0x25: 'LL_POWER_CHANGE_IND',
    0x26: 'LL_SUBRATE_REQ',
    0x27: 'LL_SUBRATE_IND',
    0x28: 'LL_CHANNEL_REPORTING_IND',
    0x29: 'LL_CHANNEL_STATUS_IND',
}


class BTLE_EMPTY_PDU(Packet):
    name = "Empty data PDU"


class BTLE_CTRL(Packet):
    name = "BTLE_CTRL"
    fields_desc = [
        ByteEnumField("opcode", 0, BTLE_BTLE_CTRL_opcode)
    ]


class LL_CONNECTION_UPDATE_IND(Packet):
    name = 'LL_CONNECTION_UPDATE_IND'
    fields_desc = [
        XByteField("win_size", 0),
        XLEShortField("win_offset", 0),
        XLEShortField("interval", 6),
        XLEShortField("latency", 0),
        XLEShortField("timeout", 50),
        XLEShortField("instant", 6),
    ]


class LL_CHANNEL_MAP_IND(Packet):
    name = 'LL_CHANNEL_MAP_IND'
    fields_desc = [
        BTLEChanMapField("chM", 0xFFFFFFFFFE),
        XLEShortField("instant", 0),
    ]


class LL_TERMINATE_IND(Packet):
    name = 'LL_TERMINATE_IND'
    fields_desc = [
        XByteField("code", 0x0),
    ]


class LL_ENC_REQ(Packet):
    name = 'LL_ENC_REQ'
    fields_desc = [
        XLELongField("rand", 0),
        XLEShortField("ediv", 0),
        XLELongField("skdm", 0),
        XLEIntField("ivm", 0),
    ]


class LL_ENC_RSP(Packet):
    name = 'LL_ENC_RSP'
    fields_desc = [
        XLELongField("skds", 0),
        XLEIntField("ivs", 0),
    ]


class LL_START_ENC_REQ(Packet):
    name = 'LL_START_ENC_REQ'
    fields_desc = []


class LL_START_ENC_RSP(Packet):
    name = 'LL_START_ENC_RSP'


class LL_UNKNOWN_RSP(Packet):
    name = 'LL_UNKNOWN_RSP'
    fields_desc = [
        XByteField("code", 0x0),
    ]


class LL_FEATURE_REQ(Packet):
    name = "LL_FEATURE_REQ"
    fields_desc = [
        BTLEFeatureField("feature_set", 0)
    ]


class LL_FEATURE_RSP(Packet):
    name = "LL_FEATURE_RSP"
    fields_desc = [
        BTLEFeatureField("feature_set", 0)
    ]


class LL_PAUSE_ENC_REQ(Packet):
    name = "LL_PAUSE_ENC_REQ"


class LL_PAUSE_ENC_RSP(Packet):
    name = "LL_PAUSE_ENC_RSP"


class LL_VERSION_IND(Packet):
    name = "LL_VERSION_IND"
    fields_desc = [
        ByteEnumField("version", 8, BTLE_Versions),
        LEShortEnumField("company", 0, BTLE_Corp_IDs),
        XLEShortField("subversion", 0)
    ]


class LL_REJECT_IND(Packet):
    name = "LL_REJECT_IND"
    fields_desc = [
        XByteField("code", 0x0),
    ]


class LL_SLAVE_FEATURE_REQ(Packet):
    name = "LL_SLAVE_FEATURE_REQ"
    fields_desc = [
        BTLEFeatureField("feature_set", 0)
    ]


class LL_CONNECTION_PARAM_REQ(Packet):
    name = "LL_CONNECTION_PARAM_REQ"
    fields_desc = [
        XShortField("interval_min", 0x6),
        XShortField("interval_max", 0x6),
        XShortField("latency", 0x0),
        XShortField("timeout", 0x0),
        XByteField("preferred_periodicity", 0x0),
        XShortField("reference_conn_evt_count", 0x0),
        XShortField("offset0", 0x0),
        XShortField("offset1", 0x0),
        XShortField("offset2", 0x0),
        XShortField("offset3", 0x0),
        XShortField("offset4", 0x0),
        XShortField("offset5", 0x0),
    ]


class LL_CONNECTION_PARAM_RSP(Packet):
    name = "LL_CONNECTION_PARAM_RSP"
    fields_desc = [
        XShortField("interval_min", 0x6),
        XShortField("interval_max", 0x6),
        XShortField("latency", 0x0),
        XShortField("timeout", 0x0),
        XByteField("preferred_periodicity", 0x0),
        XShortField("reference_conn_evt_count", 0x0),
        XShortField("offset0", 0x0),
        XShortField("offset1", 0x0),
        XShortField("offset2", 0x0),
        XShortField("offset3", 0x0),
        XShortField("offset4", 0x0),
        XShortField("offset5", 0x0),
    ]


class LL_REJECT_EXT_IND(Packet):
    name = "LL_REJECT_EXT_IND"
    fields_desc = [
        XByteField("reject_opcode", 0x0),
        XByteField("error_code", 0x0),
    ]


class LL_PING_REQ(Packet):
    name = "LL_PING_REQ"


class LL_PING_RSP(Packet):
    name = "LL_PING_RSP"


class LL_LENGTH_REQ(Packet):
    name = ' LL_LENGTH_REQ'
    fields_desc = [
        XLEShortField("max_rx_bytes", 251),
        XLEShortField("max_rx_time", 2120),
        XLEShortField("max_tx_bytes", 251),
        XLEShortField("max_tx_time", 2120),
    ]


class LL_LENGTH_RSP(Packet):
    name = ' LL_LENGTH_RSP'
    fields_desc = [
        XLEShortField("max_rx_bytes", 251),
        XLEShortField("max_rx_time", 2120),
        XLEShortField("max_tx_bytes", 251),
        XLEShortField("max_tx_time", 2120),
    ]


class LL_PHY_REQ(Packet):
    name = "LL_PHY_REQ"
    fields_desc = [
        BTLEPhysField('tx_phys', 0),
        BTLEPhysField('rx_phys', 0),
    ]


class LL_PHY_RSP(Packet):
    name = "LL_PHY_RSP"
    fields_desc = [
        BTLEPhysField('tx_phys', 0),
        BTLEPhysField('rx_phys', 0),
    ]


class LL_PHY_UPDATE_IND(Packet):
    name = "LL_PHY_UPDATE_IND"
    fields_desc = [
        BTLEPhysField('tx_phy', 0),
        BTLEPhysField('rx_phy', 0),
        XShortField("instant", 0x0),
    ]


class LL_MIN_USED_CHANNELS_IND(Packet):
    name = "LL_MIN_USED_CHANNELS_IND"
    fields_desc = [
        BTLEPhysField('phys', 0),
        ByteField("min_used_channels", 2),
    ]


class LL_CTE_REQ(Packet):
    name = "LL_CTE_REQ"
    fields_desc = [
        LEBitField('min_cte_len_req', 0, 5),
        LEBitField('rfu', 0, 1),
        LEBitField("cte_type_req", 0, 2)
    ]


class LL_CTE_RSP(Packet):
    name = "LL_CTE_RSP"
    fields_desc = []


class LL_PERIODIC_SYNC_IND(Packet):
    name = "LL_PERIODIC_SYNC_IND"
    fields_desc = [
        XLEShortField("id", 251),
        LEBitField("sync_info", 0, 18 * 8),
        XLEShortField("conn_event_count", 0),
        XLEShortField("last_pa_event_counter", 0),
        LEBitField('sid', 0, 4),
        LEBitField('a_type', 0, 1),
        LEBitField('sca', 0, 3),
        BTLEPhysField('phy', 0),
        BDAddrField("AdvA", None),
        XLEShortField("sync_conn_event_count", 0),
    ]


class LL_CLOCK_ACCURACY_REQ(Packet):
    name = "LL_CLOCK_ACCURACY_REQ"
    fields_desc = [
        XByteField("sca", 0),
    ]


class LL_CLOCK_ACCURACY_RSP(Packet):
    name = "LL_CLOCK_ACCURACY_RSP"
    fields_desc = [
        XByteField("sca", 0),
    ]


class LL_CIS_REQ(Packet):
    name = 'LL_CIS_REQ'
    fields_desc = [
        XByteField("cig_id", 0),
        XByteField("cis_id", 0),
        BTLEPhysField('phy_c_to_p', 0),
        BTLEPhysField('phy_p_to_c', 0),
        LEBitField('max_sdu_c_to_p', 0, 12),
        LEBitField('rfu1', 0, 3),
        LEBitField('framed', 0, 1),
        LEBitField('max_sdu_p_to_c', 0, 12),
        LEBitField('rfu2', 0, 4),
        LEBitField('sdu_interval_c_to_p', 0, 20),
        LEBitField('rfu3', 0, 4),
        LEBitField('sdu_interval_p_to_c', 0, 20),
        LEBitField('rfu4', 0, 4),
        XLEShortField("max_pdu_c_to_p", 0),
        XLEShortField("max_pdu_p_to_c", 0),
        XByteField("nse", 0),
        X3BytesField("subinterval", 0x0),
        LEBitField('bn_c_to_p', 0, 4),
        LEBitField('bn_p_to_c', 0, 4),
        ByteField("ft_c_to_p", 0),
        ByteField("ft_p_to_c", 0),
        XLEShortField("iso_interval", 0),
        X3BytesField("cis_offset_min", 0x0),
        X3BytesField("cis_offset_max", 0x0),
        XLEShortField("conn_event_count", 0),
    ]


class LL_CIS_RSP(Packet):
    name = 'LL_CIS_RSP'
    fields_desc = [
        X3BytesField("cis_offset_min", 0x0),
        X3BytesField("cis_offset_max", 0x0),
        XLEShortField("conn_event_count", 0),
    ]


class LL_CIS_IND(Packet):
    name = 'LL_CIS_IND'
    fields_desc = [
        XIntField("AA", 0x00),
        X3BytesField("cis_offset", 0x0),
        X3BytesField("cig_sync_delay", 0x0),
        X3BytesField("cis_sync_delay", 0x0),
        XLEShortField("conn_event_count", 0),
    ]


class LL_CIS_TERMINATE_IND(Packet):
    name = 'LL_CIS_TERMINATE_IND'
    fields_desc = [
        ByteField("cig_id", 0x0),
        ByteField("cis_id", 0x0),
        ByteField("error_code", 0x0),
    ]


class LL_POWER_CONTROL_REQ(Packet):
    name = 'LL_POWER_CONTROL_REQ'
    fields_desc = [
        ByteField("phy", 0x0),
        SignedByteField("delta", 0x0),
        SignedByteField("tx_power", 0x0),
    ]


class LL_POWER_CONTROL_RSP(Packet):
    name = 'LL_POWER_CONTROL_RSP'
    fields_desc = [
        LEBitField("min", 0, 1),
        LEBitField("max", 0, 1),
        LEBitField("rfu", 0, 6),
        SignedByteField("delta", 0),
        SignedByteField("tx_power", 0x0),
        ByteField("apr", 0x0),
    ]


class LL_POWER_CHANGE_IND(Packet):
    name = 'LL_POWER_CHANGE_IND'
    fields_desc = [
        ByteField("phy", 0x0),
        LEBitField("min", 0, 1),
        LEBitField("max", 0, 1),
        LEBitField("rfu", 0, 6),
        SignedByteField("delta", 0),
        ByteField("tx_power", 0x0),
    ]


class LL_SUBRATE_REQ(Packet):
    name = 'LL_SUBRATE_REQ'
    fields_desc = [
        LEShortField("subrate_factor_min", 0x0),
        LEShortField("subrate_factor_max", 0x0),
        LEShortField("max_latency", 0x0),
        LEShortField("continuation_number", 0x0),
        LEShortField("timeout", 0x0),
    ]


class LL_SUBRATE_IND(Packet):
    name = 'LL_SUBRATE_IND'
    fields_desc = [
        LEShortField("subrate_factor", 0x0),
        LEShortField("subrate_base_event", 0x0),
        LEShortField("latency", 0x0),
        LEShortField("continuation_number", 0x0),
        LEShortField("timeout", 0x0),
    ]


class LL_CHANNEL_REPORTING_IND(Packet):
    name = 'LL_SUBRATE_IND'
    fields_desc = [
        ByteField("enable", 0x0),
        ByteField("min_spacing", 0x0),
        ByteField("max_delay", 0x0),
    ]


class LL_CHANNEL_STATUS_IND(Packet):
    name = 'LL_CHANNEL_STATUS_IND'
    fields_desc = [
        LEBitField("channel_classification", 0, 10 * 8),
    ]


# Advertisement (37-39) channel PDUs
bind_layers(BTLE, BTLE_ADV, access_addr=0x8E89BED6)
bind_layers(BTLE, BTLE_DATA)
bind_layers(BTLE_ADV, BTLE_ADV_IND, PDU_type=0)
bind_layers(BTLE_ADV, BTLE_ADV_DIRECT_IND, PDU_type=1)
bind_layers(BTLE_ADV, BTLE_ADV_NONCONN_IND, PDU_type=2)
bind_layers(BTLE_ADV, BTLE_SCAN_REQ, PDU_type=3)
bind_layers(BTLE_ADV, BTLE_SCAN_RSP, PDU_type=4)
bind_layers(BTLE_ADV, BTLE_CONNECT_REQ, PDU_type=5)
bind_layers(BTLE_ADV, BTLE_ADV_SCAN_IND, PDU_type=6)

# Data channel (0-36) PDUs
# LLID=1 -> Continue
bind_layers(BTLE_DATA, L2CAP_Hdr, LLID=2)
bind_layers(BTLE_DATA, BTLE_CTRL, LLID=3)
bind_layers(BTLE_DATA, BTLE_EMPTY_PDU, {'len': 0, 'LLID': 1})
bind_layers(BTLE_CTRL, LL_CONNECTION_UPDATE_IND, opcode=0x00)
bind_layers(BTLE_CTRL, LL_CHANNEL_MAP_IND, opcode=0x01)
bind_layers(BTLE_CTRL, LL_TERMINATE_IND, opcode=0x02)
bind_layers(BTLE_CTRL, LL_ENC_REQ, opcode=0x03)
bind_layers(BTLE_CTRL, LL_ENC_RSP, opcode=0x04)
bind_layers(BTLE_CTRL, LL_START_ENC_REQ, opcode=0x05)
bind_layers(BTLE_CTRL, LL_START_ENC_RSP, opcode=0x06)
bind_layers(BTLE_CTRL, LL_UNKNOWN_RSP, opcode=0x07)
bind_layers(BTLE_CTRL, LL_FEATURE_REQ, opcode=0x08)
bind_layers(BTLE_CTRL, LL_FEATURE_RSP, opcode=0x09)
bind_layers(BTLE_CTRL, LL_PAUSE_ENC_REQ, opcode=0x0A)
bind_layers(BTLE_CTRL, LL_PAUSE_ENC_RSP, opcode=0x0B)
bind_layers(BTLE_CTRL, LL_VERSION_IND, opcode=0x0C)
bind_layers(BTLE_CTRL, LL_REJECT_IND, opcode=0x0D)
bind_layers(BTLE_CTRL, LL_SLAVE_FEATURE_REQ, opcode=0x0E)
bind_layers(BTLE_CTRL, LL_CONNECTION_PARAM_REQ, opcode=0x0F)
bind_layers(BTLE_CTRL, LL_CONNECTION_PARAM_RSP, opcode=0x10)
bind_layers(BTLE_CTRL, LL_REJECT_EXT_IND, opcode=0x11)
bind_layers(BTLE_CTRL, LL_PING_REQ, opcode=0x12)
bind_layers(BTLE_CTRL, LL_PING_RSP, opcode=0x13)
bind_layers(BTLE_CTRL, LL_LENGTH_REQ, opcode=0x14)
bind_layers(BTLE_CTRL, LL_LENGTH_RSP, opcode=0x15)
bind_layers(BTLE_CTRL, LL_PHY_REQ, opcode=0x16)
bind_layers(BTLE_CTRL, LL_PHY_RSP, opcode=0x17)
bind_layers(BTLE_CTRL, LL_PHY_UPDATE_IND, opcode=0x18)
bind_layers(BTLE_CTRL, LL_MIN_USED_CHANNELS_IND, opcode=0x19)
bind_layers(BTLE_CTRL, LL_CTE_REQ, opcode=0x1A)
bind_layers(BTLE_CTRL, LL_CTE_RSP, opcode=0x1B)
bind_layers(BTLE_CTRL, LL_PERIODIC_SYNC_IND, opcode=0x1C)
bind_layers(BTLE_CTRL, LL_CLOCK_ACCURACY_REQ, opcode=0x1D)
bind_layers(BTLE_CTRL, LL_CLOCK_ACCURACY_RSP, opcode=0x1E)
bind_layers(BTLE_CTRL, LL_CIS_REQ, opcode=0x1F)
bind_layers(BTLE_CTRL, LL_CIS_RSP, opcode=0x20)
bind_layers(BTLE_CTRL, LL_CIS_IND, opcode=0x21)
bind_layers(BTLE_CTRL, LL_CIS_TERMINATE_IND, opcode=0x22)
bind_layers(BTLE_CTRL, LL_POWER_CONTROL_REQ, opcode=0x23)
bind_layers(BTLE_CTRL, LL_POWER_CONTROL_RSP, opcode=0x24)
bind_layers(BTLE_CTRL, LL_POWER_CHANGE_IND, opcode=0x25)
bind_layers(BTLE_CTRL, LL_SUBRATE_REQ, opcode=0x26)
bind_layers(BTLE_CTRL, LL_SUBRATE_IND, opcode=0x27)
bind_layers(BTLE_CTRL, LL_CHANNEL_REPORTING_IND, opcode=0x28)
bind_layers(BTLE_CTRL, LL_CHANNEL_STATUS_IND, opcode=0x29)


conf.l2types.register(DLT_BLUETOOTH_LE_LL, BTLE)
conf.l2types.register(DLT_BLUETOOTH_LE_LL_WITH_PHDR, BTLE_RF)

bind_layers(BTLE_RF, BTLE)

bind_layers(PPI_Hdr, BTLE_PPI, pfh_type=PPI_BTLE)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Leonard Crestez <cdleonard@gmail.com>

# scapy.contrib.description = TCP-AO Signature Calculation
# scapy.contrib.status = loads

"""Packet-processing utilities implementing RFC5925 and RFC5926"""

import logging
from scapy.compat import orb
from scapy.layers.inet import IP, TCP
from scapy.layers.inet import tcp_pseudoheader
from scapy.layers.inet6 import IPv6
from scapy.packet import Packet
from scapy.pton_ntop import inet_pton
import socket
import struct

from typing import (
    Union,
)

logger = logging.getLogger(__name__)


def _hmac_sha1_digest(key, msg):
    # type: (bytes, bytes) -> bytes
    import hmac
    import hashlib

    return hmac.new(key, msg, hashlib.sha1).digest()


def _cmac_aes_digest(key, msg):
    # type: (bytes, bytes) -> bytes
    from cryptography.hazmat.primitives import cmac
    from cryptography.hazmat.primitives.ciphers import algorithms
    from cryptography.hazmat.backends import default_backend

    backend = default_backend()
    c = cmac.CMAC(algorithms.AES(key), backend=backend)
    c.update(bytes(msg))
    return c.finalize()


class TCPAOAlg:
    @classmethod
    def kdf(cls, master_key, context):
        # type: (bytes, bytes) -> bytes
        raise NotImplementedError()

    @classmethod
    def mac(cls, traffic_key, context):
        # type: (bytes, bytes) -> bytes
        raise NotImplementedError()

    maclen = -1


class TCPAOAlg_HMAC_SHA1(TCPAOAlg):
    @classmethod
    def kdf(cls, master_key, context):
        # type: (bytes, bytes) -> bytes
        input = b"\x01" + b"TCP-AO" + context + b"\x00\xa0"
        return _hmac_sha1_digest(master_key, input)

    @classmethod
    def mac(cls, traffic_key, message):
        # type: (bytes, bytes) -> bytes
        return _hmac_sha1_digest(traffic_key, message)[:12]

    maclen = 12


class TCPAOAlg_CMAC_AES(TCPAOAlg):
    @classmethod
    def kdf(self, master_key, context):
        # type: (bytes, bytes) -> bytes
        if len(master_key) == 16:
            key = master_key
        else:
            key = _cmac_aes_digest(b"\x00" * 16, master_key)
        return _cmac_aes_digest(key, b"\x01TCP-AO" + context + b"\x00\x80")

    @classmethod
    def mac(self, traffic_key, message):
        # type: (bytes, bytes) -> bytes
        return _cmac_aes_digest(traffic_key, message)[:12]

    maclen = 12


def get_alg(name):
    # type: (str) -> TCPAOAlg
    if name.upper() == "HMAC-SHA-1-96":
        return TCPAOAlg_HMAC_SHA1()
    elif name.upper() == "AES-128-CMAC-96":
        return TCPAOAlg_CMAC_AES()
    else:
        raise ValueError("Bad TCP AuthOpt algorithms {}".format(name))


def _get_ipvx_src(u):
    # type: (Union[IP, IPv6]) -> bytes
    if isinstance(u, IP):
        return inet_pton(socket.AF_INET, u.src)
    elif isinstance(u, IPv6):
        return inet_pton(socket.AF_INET6, u.src)
    else:
        raise Exception("Neither IP nor IPv6 found on packet")


def _get_ipvx_dst(u):
    # type: (Union[IP, IPv6]) -> bytes
    if isinstance(u, IP):
        return inet_pton(socket.AF_INET, u.dst)
    elif isinstance(u, IPv6):
        return inet_pton(socket.AF_INET6, u.dst)
    else:
        raise Exception("Neither IP nor IPv6 found on packet")


def build_context(
    saddr,  # type: bytes
    daddr,  # type: bytes
    sport,  # type: int
    dport,  # type: int
    src_isn,  # type: int
    dst_isn,  # type: int
):
    # type: (...) -> bytes
    """Build context bytes as specified by RFC5925 section 5.2"""
    if len(saddr) != len(daddr) or (len(saddr) != 4 and len(saddr) != 16):
        raise ValueError("saddr and daddr must be 4-byte or 16-byte addresses")
    return (
        saddr +
        daddr +
        struct.pack(
            "!HHII",
            sport,
            dport,
            src_isn,
            dst_isn,
        )
    )


def build_context_from_packet(
    p,  # type: Packet
    src_isn,  # type: int
    dst_isn,  # type: int
):
    # type: (...) -> bytes
    """Build context bytes as specified by RFC5925 section 5.2"""
    tcp = p[TCP]
    return build_context(
        _get_ipvx_src(tcp.underlayer),
        _get_ipvx_dst(tcp.underlayer),
        tcp.sport,
        tcp.dport,
        src_isn,
        dst_isn,
    )


def build_message_from_packet(p, include_options=True, sne=0):
    # type: (Packet, bool, int) -> bytes
    """Build message bytes as described by RFC5925 section 5.1"""
    result = bytearray()
    result += struct.pack("!I", sne)
    result += tcp_pseudoheader(p[TCP])

    # tcp header with checksum set to zero
    th_bytes = bytes(p[TCP])
    result += th_bytes[:16]
    result += b"\x00\x00"
    result += th_bytes[18:20]

    # Even if include_options=False the TCP-AO option itself is still included
    # with the MAC set to all-zeros. This means we need to parse TCP options.
    pos = 20
    th = p[TCP]
    doff = th.dataofs
    if doff is None:
        opt_len = len(th.get_field("options").i2m(th, th.options))
        doff = 5 + ((opt_len + 3) // 4)
    tcphdr_optend = doff * 4
    while pos < tcphdr_optend:
        optnum = orb(th_bytes[pos])
        pos += 1
        if optnum == 0 or optnum == 1:
            if include_options:
                result += bytearray([optnum])
            continue

        optlen = orb(th_bytes[pos])
        pos += 1
        if pos + optlen - 2 > tcphdr_optend:
            logger.info("bad tcp option %d optlen %d beyond end-of-header",
                        optnum, optlen)
            break
        if optlen < 2:
            logger.info("bad tcp option %d optlen %d less than two",
                        optnum, optlen)
            break
        if optnum == 29:
            if optlen < 4:
                logger.info("bad tcp option %d optlen %d", optnum, optlen)
                break
            result += th_bytes[pos - 2: pos + 2]
            result += (optlen - 4) * b"\x00"
        elif include_options:
            result += th_bytes[pos - 2: pos + optlen - 2]
        pos += optlen - 2
    result += bytes(p[TCP].payload)
    return result


def calc_tcpao_traffic_key(p, alg, master_key, sisn, disn):
    # type: (Packet, TCPAOAlg, bytes, int, int) -> bytes
    """Calculate TCP-AO traffic-key from packet and initial sequence numbers

    This is constant for an established connection.
    """
    return alg.kdf(master_key, build_context_from_packet(p, sisn, disn))


def calc_tcpao_mac(p, alg, traffic_key, include_options=True, sne=0):
    # type: (Packet, TCPAOAlg, bytes, bool, int) -> bytes
    """Calculate TCP-AO MAC from packet and traffic key"""
    return alg.mac(traffic_key, build_message_from_packet(
        p, include_options=include_options, sne=sne
    ))


def sign_tcpao(
    p,
    alg,
    traffic_key,
    keyid=0,
    rnextkeyid=0,
    include_options=True,
    sne=0,
):
    # type: (Packet, TCPAOAlg, bytes, int, int, bool, int) -> None
    """Calculate TCP-AO option value and insert into packet"""
    th = p[TCP]
    keyids = struct.pack("BB", keyid, rnextkeyid)
    th.options = th.options + [('AO', keyids + alg.maclen * b"\x00")]
    message_bytes = calc_tcpao_mac(
        p, alg, traffic_key, include_options=include_options, sne=sne)
    mac = alg.mac(traffic_key, message_bytes)
    th.options[-1] = ('AO', keyids + mac)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
Direct Access dictionary.
"""

from scapy.error import Scapy_Exception
from scapy.compat import plain_str

# Typing
from typing import (
    Any,
    Dict,
    Generic,
    Iterator,
    List,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from scapy.compat import Self


###############################
#  Direct Access dictionary   #
###############################


def fixname(x):
    # type: (Union[bytes, str]) -> str
    """
    Modifies a string to make sure it can be used as an attribute name.
    """
    x = plain_str(x)
    if x and str(x[0]) in "0123456789":
        x = "n_" + x
    return x.translate(
        "________________________________________________"
        "0123456789_______ABCDEFGHIJKLMNOPQRSTUVWXYZ______"
        "abcdefghijklmnopqrstuvwxyz____________________________"
        "______________________________________________________"
        "___________________________________________________"
    )


class DADict_Exception(Scapy_Exception):
    pass


_K = TypeVar('_K')  # Key type
_V = TypeVar('_V')  # Value type


class DADict(Generic[_K, _V]):
    """
    Direct Access Dictionary

    This acts like a dict, but it provides a direct attribute access
    to its keys through its values. This is used to store protocols,
    manuf...

    For instance, scapy fields will use a DADict as an enum::

        ETHER_TYPES[2048] -> IPv4

    Whereas humans can access::

        ETHER_TYPES.IPv4 -> 2048
    """
    __slots__ = ["_name", "d"]

    def __init__(self, _name="DADict", **kargs):
        # type: (str, **Any) -> None
        self._name = _name
        self.d = {}  # type: Dict[_K, _V]
        self.update(kargs)  # type: ignore

    def ident(self, v):
        # type: (_V) -> str
        """
        Return value that is used as key for the direct access
        """
        if isinstance(v, (str, bytes)):
            return fixname(v)
        return "unknown"

    def update(self, *args, **kwargs):
        # type: (*Dict[_K, _V], **Dict[_K, _V]) -> None
        for k, v in dict(*args, **kwargs).items():
            self[k] = v  # type: ignore

    def iterkeys(self):
        # type: () -> Iterator[_K]
        for x in self.d:
            if not isinstance(x, str) or x[0] != "_":
                yield x

    def keys(self):
        # type: () -> List[_K]
        return list(self.iterkeys())

    def __iter__(self):
        # type: () -> Iterator[_K]
        return self.iterkeys()

    def itervalues(self):
        # type: () -> Iterator[_V]
        return self.d.values()  # type: ignore

    def values(self):
        # type: () -> List[_V]
        return list(self.itervalues())

    def _show(self):
        # type: () -> None
        for k in self.iterkeys():
            print("%10s = %r" % (k, self[k]))

    def __repr__(self):
        # type: () -> str
        return "<%s - %s elements>" % (self._name, len(self))

    def __getitem__(self, attr):
        # type: (_K) -> _V
        return self.d[attr]

    def __setitem__(self, attr, val):
        # type: (_K, _V) -> None
        self.d[attr] = val

    def __len__(self):
        # type: () -> int
        return len(self.d)

    def __nonzero__(self):
        # type: () -> bool
        # Always has at least its name
        return len(self) > 1
    __bool__ = __nonzero__

    def __getattr__(self, attr):
        # type: (str) -> _K
        try:
            return object.__getattribute__(self, attr)  # type: ignore
        except AttributeError:
            for k, v in self.d.items():
                if self.ident(v) == attr:
                    return k
        raise AttributeError

    def __dir__(self):
        # type: () -> List[str]
        return [self.ident(x) for x in self.itervalues()]

    def __reduce__(self):
        # type: () -> Tuple[Type[Self], Tuple[str], Tuple[Dict[_K, _V]]]
        return (self.__class__, (self._name,), (self.d,))

    def __setstate__(self, state):
        # type: (Tuple[Dict[_K, _V]]) -> Self
        self.d.update(state[0])
        return self



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
Package holding ASN.1 related modules.
"""



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.description = Staged AutomotiveTestCase base classes
# scapy.contrib.status = library


from scapy.contrib.automotive import log_automotive
from scapy.contrib.automotive.scanner.graph import _Edge
from scapy.contrib.automotive.ecu import EcuState, EcuResponse, Ecu
from scapy.contrib.automotive.scanner.test_case import AutomotiveTestCaseABC, \
    TestCaseGenerator, StateGenerator, _SocketUnion

# Typing imports
from typing import (
    Any,
    List,
    Optional,
    Dict,
    Callable,
    cast,
    Tuple,
    TYPE_CHECKING,
)
if TYPE_CHECKING:
    from scapy.contrib.automotive.scanner.test_case import _TransitionTuple
    from scapy.contrib.automotive.scanner.configuration import \
        AutomotiveTestCaseExecutorConfiguration

# type definitions
_TestCaseConnectorCallable = \
    Callable[[AutomotiveTestCaseABC, AutomotiveTestCaseABC], Dict[str, Any]]


class StagedAutomotiveTestCase(AutomotiveTestCaseABC, TestCaseGenerator, StateGenerator):  # noqa: E501
    """ Helper object to build a pipeline of TestCases. This allows to combine
    TestCases and to execute them after each other. Custom connector functions
    can be used to exchange and manipulate the configuration of a subsequent
    TestCase.

    :param test_cases: A list of objects following the AutomotiveTestCaseABC
        interface
    :param connectors: A list of connector functions. A connector function
        takes two TestCase objects and returns a dictionary which is provided
        to the second TestCase as kwargs of the execute function.


    Example:
        >>> class MyTestCase2(AutomotiveTestCaseABC):
        >>>     pass
        >>>
        >>> class MyTestCase1(AutomotiveTestCaseABC):
        >>>     pass
        >>>
        >>> def connector(testcase1, testcase2):
        >>>     scan_range = len(testcase1.results)
        >>>     return {"verbose": True, "scan_range": scan_range}
        >>>
        >>> tc1 = MyTestCase1()
        >>> tc2 = MyTestCase2()
        >>> pipeline = StagedAutomotiveTestCase([tc1, tc2], [None, connector])
    """

    # Delay the increment of a stage after the current stage is finished
    # has_completed() has to be called five times in order to increment the
    # current stage. This ensures, that the current stage is executed for
    # all possible states of the DUT, and no state is missed for the first
    # TestCase.
    __delay_stages = 5

    def __init__(self,
                 test_cases,  # type: List[AutomotiveTestCaseABC]
                 connectors=None  # type: Optional[List[Optional[_TestCaseConnectorCallable]]]  # noqa: E501
                 ):  # type: (...) -> None
        super(StagedAutomotiveTestCase, self).__init__()
        self.__test_cases = test_cases
        self.__connectors = connectors
        self.__stage_index = 0
        self.__completion_delay = 0
        self.__current_kwargs = None  # type: Optional[Dict[str, Any]]

    def __getitem__(self, item):
        # type: (int) -> AutomotiveTestCaseABC
        return self.__test_cases[item]

    def __len__(self):
        # type: () -> int
        return len(self.__test_cases)

    # TODO: Fix unit tests and remove this function
    def __reduce__(self):  # type: ignore
        f, t, d = super(StagedAutomotiveTestCase, self).__reduce__()  # type: ignore  # noqa: E501
        try:
            del d["_StagedAutomotiveTestCase__connectors"]
        except KeyError:
            pass
        return f, t, d

    @property
    def test_cases(self):
        # type: () -> List[AutomotiveTestCaseABC]
        return self.__test_cases

    @property
    def current_test_case(self):
        # type: () -> AutomotiveTestCaseABC
        return self[self.__stage_index]

    @property
    def current_connector(self):
        # type: () -> Optional[_TestCaseConnectorCallable]
        if not self.__connectors:
            return None
        else:
            return self.__connectors[self.__stage_index]

    @property
    def previous_test_case(self):
        # type: () -> Optional[AutomotiveTestCaseABC]
        return self.__test_cases[self.__stage_index - 1] if \
            self.__stage_index > 0 else None

    def get_generated_test_case(self):
        # type: () -> Optional[AutomotiveTestCaseABC]
        try:
            test_case = cast(TestCaseGenerator, self.current_test_case)
            return test_case.get_generated_test_case()
        except AttributeError:
            return None

    def get_new_edge(self,
                     socket,  # type: _SocketUnion
                     config  # type: AutomotiveTestCaseExecutorConfiguration
                     ):  # type: (...) -> Optional[_Edge]
        try:
            test_case = cast(StateGenerator, self.current_test_case)
            return test_case.get_new_edge(socket, config)
        except AttributeError:
            return None

    def get_transition_function(self, socket, edge):
        # type: (_SocketUnion, _Edge) -> Optional[_TransitionTuple]
        try:
            test_case = cast(StateGenerator, self.current_test_case)
            return test_case.get_transition_function(socket, edge)
        except AttributeError:
            return None

    def has_completed(self, state):
        # type: (EcuState) -> bool
        if not (self.current_test_case.has_completed(state) and
                self.current_test_case.completed):
            # current test_case not fully completed
            # reset completion delay, since new states could have been appeared
            self.__completion_delay = 0
            return False

        # current stage is finished. We have to increase the stage
        if self.__completion_delay < StagedAutomotiveTestCase.__delay_stages:
            # First we wait five more iteration of the executor
            # Maybe one more execution reveals new states of other
            # test_cases
            self.__completion_delay += 1
            return False

        # current test_case is fully completed
        elif self.__stage_index == len(self.__test_cases) - 1:
            # this test_case was the last test_case... nothing to do
            return True

        else:
            # We waited more iterations and no new state appeared,
            # let's enter the next stage
            log_automotive.info(
                "Staged AutomotiveTestCase %s completed",
                self.current_test_case.__class__.__name__)
            self.__stage_index += 1
            self.__completion_delay = 0
        return False

    def pre_execute(self,
                    socket,  # type: _SocketUnion
                    state,  # type: EcuState
                    global_configuration  # type: AutomotiveTestCaseExecutorConfiguration  # noqa: E501
                    ):  # type: (...) -> None
        test_case_cls = self.current_test_case.__class__
        try:
            self.__current_kwargs = global_configuration[
                test_case_cls.__name__]
        except KeyError:
            self.__current_kwargs = dict()
            global_configuration[test_case_cls.__name__] = \
                self.__current_kwargs

        if callable(self.current_connector) and self.__stage_index > 0:
            if self.previous_test_case:
                con = self.current_connector  # type: _TestCaseConnectorCallable  # noqa: E501
                con_kwargs = con(self.previous_test_case,
                                 self.current_test_case)
                if self.__current_kwargs is not None and con_kwargs is not None:  # noqa: E501
                    self.__current_kwargs.update(con_kwargs)

        log_automotive.debug("Stage AutomotiveTestCase %s kwargs: %s",
                             self.current_test_case.__class__.__name__,
                             self.__current_kwargs)

        self.current_test_case.pre_execute(socket, state, global_configuration)

    def execute(self, socket, state, **kwargs):
        # type: (_SocketUnion, EcuState, Any) -> None
        kwargs.update(self.__current_kwargs or dict())
        self.current_test_case.execute(socket, state, **kwargs)

    def post_execute(self,
                     socket,  # type: _SocketUnion
                     state,  # type: EcuState
                     global_configuration  # type: AutomotiveTestCaseExecutorConfiguration  # noqa: E501
                     ):  # type: (...) -> None
        self.current_test_case.post_execute(
            socket, state, global_configuration)

    @staticmethod
    def _show_headline(headline, sep="="):
        # type: (str, str) -> str
        s = "\n\n" + sep * (len(headline) + 10) + "\n"
        s += " " * 5 + headline + "\n"
        s += sep * (len(headline) + 10) + "\n"
        return s + "\n"

    def show(self, dump=False, filtered=True, verbose=False):
        # type: (bool, bool, bool) -> Optional[str]
        s = self._show_headline("AutomotiveTestCase Pipeline", "=")
        for idx, t in enumerate(self.__test_cases):
            s += self._show_headline(
                "AutomotiveTestCase Stage %d" % idx, "-")
            s += t.show(True, filtered, verbose) or ""

        if dump:
            return s + "\n"
        else:
            print(s)
            return None

    @property
    def completed(self):
        # type: () -> bool
        return all(e.completed for e in self.__test_cases) and \
            self.__completion_delay >= StagedAutomotiveTestCase.__delay_stages

    @property
    def supported_responses(self):
        # type: () -> List[EcuResponse]
        supported_responses = list()
        for tc in self.test_cases:
            supported_responses += tc.supported_responses

        supported_responses.sort(key=Ecu.sort_key_func)
        return supported_responses

    def runtime_estimation(self):
        # type: () -> Optional[Tuple[int, int, float]]

        if hasattr(self.current_test_case, "runtime_estimation"):
            cur_est = self.current_test_case.runtime_estimation()
            if cur_est:
                return len(self.test_cases), \
                    self.__stage_index, \
                    float(self.__stage_index) / len(self.test_cases) + \
                    cur_est[2] / len(self.test_cases)

        return len(self.test_cases), \
            self.__stage_index, \
            float(self.__stage_index) / len(self.test_cases)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2007, 2008, 2009 Arnaud Ebalard
#               2015, 2016, 2017 Maxence Tury
#               2019 Romain Perez

"""
TLS key exchange logic.
"""

import math
import struct

from scapy.config import conf, crypto_validator
from scapy.error import warning
from scapy.fields import ByteEnumField, ByteField, EnumField, FieldLenField, \
    FieldListField, PacketField, ShortEnumField, ShortField, \
    StrFixedLenField, StrLenField
from scapy.compat import orb
from scapy.packet import Packet, Raw, Padding
from scapy.layers.tls.cert import PubKeyRSA, PrivKeyRSA
from scapy.layers.tls.session import _GenericTLSSessionInheritance
from scapy.layers.tls.basefields import _tls_version, _TLSClientVersionField
from scapy.layers.tls.crypto.pkcs1 import pkcs_i2osp, pkcs_os2ip
from scapy.layers.tls.crypto.groups import (
    _ffdh_groups,
    _tls_named_curves,
    _tls_named_groups_generate,
    _tls_named_groups_import,
    _tls_named_groups_pubbytes,
)


if conf.crypto_valid:
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives.asymmetric import dh, ec
    from cryptography.hazmat.primitives import serialization
if conf.crypto_valid_advanced:
    from cryptography.hazmat.primitives.asymmetric import x25519
    from cryptography.hazmat.primitives.asymmetric import x448


###############################################################################
#   Common Fields                                                             #
###############################################################################

_tls_hash_sig = {0x0000: "none+anon", 0x0001: "none+rsa",
                 0x0002: "none+dsa", 0x0003: "none+ecdsa",
                 0x0100: "md5+anon", 0x0101: "md5+rsa",
                 0x0102: "md5+dsa", 0x0103: "md5+ecdsa",
                 0x0200: "sha1+anon", 0x0201: "sha1+rsa",
                 0x0202: "sha1+dsa", 0x0203: "sha1+ecdsa",
                 0x0300: "sha224+anon", 0x0301: "sha224+rsa",
                 0x0302: "sha224+dsa", 0x0303: "sha224+ecdsa",
                 0x0400: "sha256+anon", 0x0401: "sha256+rsa",
                 0x0402: "sha256+dsa", 0x0403: "sha256+ecdsa",
                 0x0500: "sha384+anon", 0x0501: "sha384+rsa",
                 0x0502: "sha384+dsa", 0x0503: "sha384+ecdsa",
                 0x0600: "sha512+anon", 0x0601: "sha512+rsa",
                 0x0602: "sha512+dsa", 0x0603: "sha512+ecdsa",
                 0x0804: "sha256+rsaepss", 0x0805: "sha384+rsaepss",
                 0x0806: "sha512+rsaepss", 0x0807: "ed25519",
                 0x0808: "ed448", 0x0809: "sha256+rsapss",
                 0x080a: "sha384+rsapss", 0x080b: "sha512+rsapss"}


def phantom_mode(pkt):
    """
    We expect this. If tls_version is not set, this means we did not process
    any complete ClientHello, so we're most probably reading/building a
    signature_algorithms extension, hence we cannot be in phantom_mode.
    However, if the tls_version has been set, we test for TLS 1.2.
    """
    if not pkt.tls_session:
        return False
    if not pkt.tls_session.tls_version:
        return False
    return pkt.tls_session.tls_version < 0x0303


def phantom_decorate(f, get_or_add):
    """
    Decorator for version-dependent fields.
    If get_or_add is True (means get), we return s, self.phantom_value.
    If it is False (means add), we return s.
    """
    def wrapper(*args):
        self, pkt, s = args[:3]
        if phantom_mode(pkt):
            if get_or_add:
                return s, self.phantom_value
            return s
        return f(*args)
    return wrapper


class SigAndHashAlgField(EnumField):
    """Used in _TLSSignature."""
    phantom_value = None
    getfield = phantom_decorate(EnumField.getfield, True)
    addfield = phantom_decorate(EnumField.addfield, False)


class SigAndHashAlgsLenField(FieldLenField):
    """Used in TLS_Ext_SignatureAlgorithms and TLSCertificateResquest."""
    phantom_value = 0
    getfield = phantom_decorate(FieldLenField.getfield, True)
    addfield = phantom_decorate(FieldLenField.addfield, False)


class SigAndHashAlgsField(FieldListField):
    """Used in TLS_Ext_SignatureAlgorithms and TLSCertificateResquest."""
    phantom_value = []
    getfield = phantom_decorate(FieldListField.getfield, True)
    addfield = phantom_decorate(FieldListField.addfield, False)


class SigLenField(FieldLenField):
    """There is a trick for SSLv2, which uses implicit lengths..."""

    def getfield(self, pkt, s):
        v = pkt.tls_session.tls_version
        if v and v < 0x0300:
            return s, None
        return super(SigLenField, self).getfield(pkt, s)

    def addfield(self, pkt, s, val):
        """With SSLv2 you will never be able to add a sig_len."""
        v = pkt.tls_session.tls_version
        if v and v < 0x0300:
            return s
        return super(SigLenField, self).addfield(pkt, s, val)


class SigValField(StrLenField):
    """There is a trick for SSLv2, which uses implicit lengths..."""

    def getfield(self, pkt, m):
        s = pkt.tls_session
        if s.tls_version and s.tls_version < 0x0300:
            if len(s.client_certs) > 0:
                sig_len = s.client_certs[0].pubkey.pubkey.key_size // 8
            else:
                warning("No client certificate provided. "
                        "We're making a wild guess about the signature size.")
                sig_len = 256
            return m[sig_len:], self.m2i(pkt, m[:sig_len])
        return super(SigValField, self).getfield(pkt, m)


class _TLSSignature(_GenericTLSSessionInheritance):
    """
    Prior to TLS 1.2, digitally-signed structure implicitly used the
    concatenation of a MD5 hash and a SHA-1 hash.
    Then TLS 1.2 introduced explicit SignatureAndHashAlgorithms,
    i.e. couples of (hash_alg, sig_alg). See RFC 5246, section 7.4.1.4.1.

    By default, the _TLSSignature implements the TLS 1.2 scheme,
    but if it is provided a TLS context with a tls_version < 0x0303
    at initialization, it will fall back to the implicit signature.
    Even more, the 'sig_len' field won't be used with SSLv2.
    """
    name = "TLS Digital Signature"
    fields_desc = [SigAndHashAlgField("sig_alg", None, _tls_hash_sig),
                   SigLenField("sig_len", None, fmt="!H",
                               length_of="sig_val"),
                   SigValField("sig_val", None,
                               length_from=lambda pkt: pkt.sig_len)]

    def __init__(self, *args, **kargs):
        super(_TLSSignature, self).__init__(*args, **kargs)
        if self.sig_alg is None and "sig_alg" not in kargs:
            # Default sig_alg
            self.sig_alg = 0x0804
            if self.tls_session and self.tls_session.tls_version:
                s = self.tls_session
                if s.selected_sig_alg:
                    self.sig_alg = s.selected_sig_alg
                elif s.tls_version < 0x0303:
                    self.sig_alg = None
                elif s.tls_version == 0x0304:
                    # For TLS 1.3 signatures, set the signature
                    # algorithm to RSA-PSS
                    self.sig_alg = 0x0804

    def post_dissection(self, r):
        # for client
        self.tls_session.selected_sig_alg = self.sig_alg

    def _update_sig(self, m, key):
        """
        Sign 'm' with the PrivKey 'key' and update our own 'sig_val'.
        Note that, even when 'sig_alg' is not None, we use the signature scheme
        of the PrivKey (neither do we care to compare the both of them).
        """
        if self.sig_alg is None:
            if self.tls_session.tls_version >= 0x0300:
                self.sig_val = key.sign(m, t='pkcs', h='md5-sha1')
            else:
                self.sig_val = key.sign(m, t='pkcs', h='md5')
        else:
            if self.sig_alg in [0x0807, 0x0808]:  # ed25519, ed448
                h, t = _tls_hash_sig[self.sig_alg], None
            else:
                h, sig = _tls_hash_sig[self.sig_alg].split('+')
                if sig.endswith('pss'):
                    t = "pss"
                else:
                    t = "pkcs"
            self.sig_val = key.sign(m, t=t, h=h)

    def _verify_sig(self, m, cert):
        """
        Verify that our own 'sig_val' carries the signature of 'm' by the
        key associated to the Cert 'cert'.
        """
        if self.sig_val:
            if self.sig_alg:
                if self.sig_alg in [0x0807, 0x0808]:  # ed25519, ed448
                    h, t = _tls_hash_sig[self.sig_alg], None
                else:
                    h, sig = _tls_hash_sig[self.sig_alg].split('+')
                    if sig.endswith('pss'):
                        t = "pss"
                    else:
                        t = "pkcs"
                return cert.verify(m, self.sig_val, t=t, h=h)
            else:
                if self.tls_session.tls_version >= 0x0300:
                    return cert.verify(m, self.sig_val, t='pkcs', h='md5-sha1')
                else:
                    return cert.verify(m, self.sig_val, t='pkcs', h='md5')
        return False

    def guess_payload_class(self, p):
        return Padding


class _TLSSignatureField(PacketField):
    """
    Used for 'digitally-signed struct' in several ServerKeyExchange,
    and also in CertificateVerify. We can handle the anonymous case.
    """
    __slots__ = ["length_from"]

    def __init__(self, name, default, length_from=None):
        self.length_from = length_from
        PacketField.__init__(self, name, default, _TLSSignature)

    def m2i(self, pkt, m):
        tmp_len = self.length_from(pkt)
        if tmp_len == 0:
            return None
        return _TLSSignature(m, tls_session=pkt.tls_session)

    def getfield(self, pkt, s):
        i = self.m2i(pkt, s)
        if i is None:
            return s, None
        remain = b""
        if conf.padding_layer in i:
            r = i[conf.padding_layer]
            del r.underlayer.payload
            remain = r.load
        return remain, i


class _TLSServerParamsField(PacketField):
    """
    This is a dispatcher for the Server*DHParams below, used in
    TLSServerKeyExchange and based on the key_exchange.server_kx_msg_cls.
    When this cls is None, it means that we should not see a ServerKeyExchange,
    so we grab everything within length_from and make it available using Raw.

    When the context has not been set (e.g. when no ServerHello was parsed or
    dissected beforehand), we (kinda) clumsily set the cls by trial and error.
    XXX We could use Serv*DHParams.check_params() once it has been implemented.
    """
    __slots__ = ["length_from"]

    def __init__(self, name, default, length_from=None):
        self.length_from = length_from
        PacketField.__init__(self, name, default, None)

    def m2i(self, pkt, m):
        s = pkt.tls_session
        tmp_len = self.length_from(pkt)
        if s.prcs:
            cls = s.prcs.key_exchange.server_kx_msg_cls(m)
            if cls is None:
                return Raw(m[:tmp_len]) / Padding(m[tmp_len:])
            return cls(m, tls_session=s)
        else:
            try:
                p = ServerDHParams(m, tls_session=s)
                if pkcs_os2ip(p.load[:2]) not in _tls_hash_sig:
                    raise Exception
                return p
            except Exception:
                cls = _tls_server_ecdh_cls_guess(m)
                p = cls(m, tls_session=s)
                if pkcs_os2ip(p.load[:2]) not in _tls_hash_sig:
                    return Raw(m[:tmp_len]) / Padding(m[tmp_len:])
                return p


###############################################################################
#   Server Key Exchange parameters & value                                    #
###############################################################################

# Finite Field Diffie-Hellman

class ServerDHParams(_GenericTLSSessionInheritance):
    """
    ServerDHParams for FFDH-based key exchanges, as defined in RFC 5246/7.4.3.

    Either with .fill_missing() or .post_dissection(), the server_kx_privkey or
    server_kx_pubkey of the TLS context are updated according to the
    parsed/assembled values. It is the user's responsibility to store and
    restore the original values if he wants to keep them. For instance, this
    could be done between the writing of a ServerKeyExchange and the receiving
    of a ClientKeyExchange (which includes secret generation).
    """
    name = "Server FFDH parameters"
    fields_desc = [FieldLenField("dh_plen", None, length_of="dh_p"),
                   StrLenField("dh_p", "",
                               length_from=lambda pkt: pkt.dh_plen),
                   FieldLenField("dh_glen", None, length_of="dh_g"),
                   StrLenField("dh_g", "",
                               length_from=lambda pkt: pkt.dh_glen),
                   FieldLenField("dh_Yslen", None, length_of="dh_Ys"),
                   StrLenField("dh_Ys", "",
                               length_from=lambda pkt: pkt.dh_Yslen)]

    @crypto_validator
    def fill_missing(self):
        """
        We do not want TLSServerKeyExchange.build() to overload and recompute
        things every time it is called. This method can be called specifically
        to have things filled in a smart fashion.

        Note that we do not expect default_params.g to be more than 0xff.
        """
        s = self.tls_session

        default_params = _ffdh_groups['modp2048'][0].parameter_numbers()
        default_mLen = _ffdh_groups['modp2048'][1]

        if not self.dh_p:
            self.dh_p = pkcs_i2osp(default_params.p, default_mLen // 8)
        if self.dh_plen is None:
            self.dh_plen = len(self.dh_p)
        s.kx_group = "ffdhe%s" % (self.dh_plen * 8)

        if not self.dh_g:
            self.dh_g = pkcs_i2osp(default_params.g, 1)
        if self.dh_glen is None:
            self.dh_glen = 1

        p = pkcs_os2ip(self.dh_p)
        g = pkcs_os2ip(self.dh_g)
        real_params = dh.DHParameterNumbers(p, g).parameters(default_backend())

        if not self.dh_Ys:
            s.server_kx_privkey = real_params.generate_private_key()
            pubkey = s.server_kx_privkey.public_key()
            y = pubkey.public_numbers().y
            self.dh_Ys = pkcs_i2osp(y, pubkey.key_size // 8)
        # else, we assume that the user wrote the server_kx_privkey by himself
        if self.dh_Yslen is None:
            self.dh_Yslen = len(self.dh_Ys)

        if not s.client_kx_ffdh_params:
            s.client_kx_ffdh_params = real_params

    @crypto_validator
    def register_pubkey(self):
        """
        XXX Check that the pubkey received is in the group.
        """
        p = pkcs_os2ip(self.dh_p)
        g = pkcs_os2ip(self.dh_g)
        pn = dh.DHParameterNumbers(p, g)

        y = pkcs_os2ip(self.dh_Ys)
        public_numbers = dh.DHPublicNumbers(y, pn)

        s = self.tls_session
        s.server_kx_pubkey = public_numbers.public_key(default_backend())
        s.kx_group = "ffdhe%s" % (self.dh_plen * 8)

        if not s.client_kx_ffdh_params:
            s.client_kx_ffdh_params = pn.parameters(default_backend())

    def post_dissection(self, r):
        try:
            self.register_pubkey()
        except ImportError:
            pass

    def guess_payload_class(self, p):
        """
        The signature after the params gets saved as Padding.
        This way, the .getfield() which _TLSServerParamsField inherits
        from PacketField will return the signature remain as expected.
        """
        return Padding


# Elliptic Curve Diffie-Hellman

_tls_ec_curve_types = {1: "explicit_prime",
                       2: "explicit_char2",
                       3: "named_curve"}

_tls_ec_basis_types = {0: "ec_basis_trinomial", 1: "ec_basis_pentanomial"}


class ECCurvePkt(Packet):
    name = "Elliptic Curve"
    fields_desc = [FieldLenField("alen", None, length_of="a", fmt="B"),
                   StrLenField("a", "", length_from=lambda pkt: pkt.alen),
                   FieldLenField("blen", None, length_of="b", fmt="B"),
                   StrLenField("b", "", length_from=lambda pkt: pkt.blen)]


# Char2 Curves

class ECTrinomialBasis(Packet):
    name = "EC Trinomial Basis"
    val = 0
    fields_desc = [FieldLenField("klen", None, length_of="k", fmt="B"),
                   StrLenField("k", "", length_from=lambda pkt: pkt.klen)]

    def guess_payload_class(self, p):
        return Padding


class ECPentanomialBasis(Packet):
    name = "EC Pentanomial Basis"
    val = 1
    fields_desc = [FieldLenField("k1len", None, length_of="k1", fmt="B"),
                   StrLenField("k1", "", length_from=lambda pkt: pkt.k1len),
                   FieldLenField("k2len", None, length_of="k2", fmt="B"),
                   StrLenField("k2", "", length_from=lambda pkt: pkt.k2len),
                   FieldLenField("k3len", None, length_of="k3", fmt="B"),
                   StrLenField("k3", "", length_from=lambda pkt: pkt.k3len)]

    def guess_payload_class(self, p):
        return Padding


_tls_ec_basis_cls = {0: ECTrinomialBasis, 1: ECPentanomialBasis}


class _ECBasisTypeField(ByteEnumField):
    __slots__ = ["basis_type_of"]

    def __init__(self, name, default, enum, basis_type_of, remain=0):
        self.basis_type_of = basis_type_of
        EnumField.__init__(self, name, default, enum, "B")

    def i2m(self, pkt, x):
        if x is None:
            fld, fval = pkt.getfield_and_val(self.basis_type_of)
            x = fld.i2basis_type(pkt, fval)
        return x


class _ECBasisField(PacketField):
    __slots__ = ["clsdict", "basis_type_from"]

    def __init__(self, name, default, basis_type_from, clsdict):
        self.clsdict = clsdict
        self.basis_type_from = basis_type_from
        PacketField.__init__(self, name, default, None)

    def m2i(self, pkt, m):
        basis = self.basis_type_from(pkt)
        cls = self.clsdict[basis]
        return cls(m)

    def i2basis_type(self, pkt, x):
        val = 0
        try:
            val = x.val
        except Exception:
            pass
        return val


# Distinct ECParameters
##
# To support the different ECParameters structures defined in Sect. 5.4 of
# RFC 4492, we define 3 separates classes for implementing the 3 associated
# ServerECDHParams: ServerECDHNamedCurveParams, ServerECDHExplicitPrimeParams
# and ServerECDHExplicitChar2Params (support for this one is only partial).
# The most frequent encounter of the 3 is (by far) ServerECDHNamedCurveParams.

class ServerECDHExplicitPrimeParams(_GenericTLSSessionInheritance):
    """
    We provide parsing abilities for ExplicitPrimeParams, but there is no
    support from the cryptography library, hence no context operations.
    """
    name = "Server ECDH parameters - Explicit Prime"
    fields_desc = [ByteEnumField("curve_type", 1, _tls_ec_curve_types),
                   FieldLenField("plen", None, length_of="p", fmt="B"),
                   StrLenField("p", "", length_from=lambda pkt: pkt.plen),
                   PacketField("curve", None, ECCurvePkt),
                   FieldLenField("baselen", None, length_of="base", fmt="B"),
                   StrLenField("base", "",
                               length_from=lambda pkt: pkt.baselen),
                   FieldLenField("orderlen", None,
                                 length_of="order", fmt="B"),
                   StrLenField("order", "",
                               length_from=lambda pkt: pkt.orderlen),
                   FieldLenField("cofactorlen", None,
                                 length_of="cofactor", fmt="B"),
                   StrLenField("cofactor", "",
                               length_from=lambda pkt: pkt.cofactorlen),
                   FieldLenField("pointlen", None,
                                 length_of="point", fmt="B"),
                   StrLenField("point", "",
                               length_from=lambda pkt: pkt.pointlen)]

    def fill_missing(self):
        """
        Note that if it is not set by the user, the cofactor will always
        be 1. It is true for most, but not all, TLS elliptic curves.
        """
        if self.curve_type is None:
            self.curve_type = _tls_ec_curve_types["explicit_prime"]

    def guess_payload_class(self, p):
        return Padding


class ServerECDHExplicitChar2Params(_GenericTLSSessionInheritance):
    """
    We provide parsing abilities for Char2Params, but there is no
    support from the cryptography library, hence no context operations.
    """
    name = "Server ECDH parameters - Explicit Char2"
    fields_desc = [ByteEnumField("curve_type", 2, _tls_ec_curve_types),
                   ShortField("m", None),
                   _ECBasisTypeField("basis_type", None,
                                     _tls_ec_basis_types, "basis"),
                   _ECBasisField("basis", ECTrinomialBasis(),
                                 lambda pkt: pkt.basis_type,
                                 _tls_ec_basis_cls),
                   PacketField("curve", ECCurvePkt(), ECCurvePkt),
                   FieldLenField("baselen", None, length_of="base", fmt="B"),
                   StrLenField("base", "",
                               length_from=lambda pkt: pkt.baselen),
                   ByteField("order", None),
                   ByteField("cofactor", None),
                   FieldLenField("pointlen", None,
                                 length_of="point", fmt="B"),
                   StrLenField("point", "",
                               length_from=lambda pkt: pkt.pointlen)]

    def fill_missing(self):
        if self.curve_type is None:
            self.curve_type = _tls_ec_curve_types["explicit_char2"]

    def guess_payload_class(self, p):
        return Padding


class ServerECDHNamedCurveParams(_GenericTLSSessionInheritance):
    name = "Server ECDH parameters - Named Curve"
    fields_desc = [ByteEnumField("curve_type", 3, _tls_ec_curve_types),
                   ShortEnumField("named_curve", None, _tls_named_curves),
                   FieldLenField("pointlen", None,
                                 length_of="point", fmt="B"),
                   StrLenField("point", None,
                               length_from=lambda pkt: pkt.pointlen)]

    @crypto_validator
    def fill_missing(self):
        """
        We do not want TLSServerKeyExchange.build() to overload and recompute
        things every time it is called. This method can be called specifically
        to have things filled in a smart fashion.

        XXX We should account for the point_format (before 'point' filling).
        """
        s = self.tls_session

        if self.curve_type is None:
            self.curve_type = _tls_ec_curve_types["named_curve"]

        if self.named_curve is None:
            self.named_curve = 23

        curve_group = self.named_curve
        if curve_group not in _tls_named_curves:
            # this fallback is arguable
            curve_group = 23  # default to secp256r1
        s.server_kx_privkey = _tls_named_groups_generate(curve_group)
        s.kx_group = _tls_named_curves.get(curve_group, str(curve_group))

        if self.point is None:
            self.point = _tls_named_groups_pubbytes(
                s.server_kx_privkey
            )

        # else, we assume that the user wrote the server_kx_privkey by himself
        if self.pointlen is None:
            self.pointlen = len(self.point)

        if not s.client_kx_ecdh_params:
            s.client_kx_ecdh_params = curve_group

    @crypto_validator
    def register_pubkey(self):
        """
        XXX Support compressed point format.
        XXX Check that the pubkey received is on the curve.
        """
        # point_format = 0
        # if self.point[0] in [b'\x02', b'\x03']:
        #    point_format = 1

        s = self.tls_session
        s.server_kx_pubkey = _tls_named_groups_import(
            self.named_curve,
            self.point
        )
        s.kx_group = _tls_named_curves.get(self.named_curve, str(self.named_curve))

        if not s.client_kx_ecdh_params:
            s.client_kx_ecdh_params = self.named_curve

    def post_dissection(self, r):
        try:
            self.register_pubkey()
        except ImportError:
            pass

    def guess_payload_class(self, p):
        return Padding


_tls_server_ecdh_cls = {1: ServerECDHExplicitPrimeParams,
                        2: ServerECDHExplicitChar2Params,
                        3: ServerECDHNamedCurveParams}


def _tls_server_ecdh_cls_guess(m):
    if not m:
        return None
    curve_type = orb(m[0])
    return _tls_server_ecdh_cls.get(curve_type, None)


# RSA Encryption (export)

class ServerRSAParams(_GenericTLSSessionInheritance):
    """
    Defined for RSA_EXPORT kx : it enables servers to share RSA keys shorter
    than their principal {>512}-bit key, when it is not allowed for kx.

    This should not appear in standard RSA kx negotiation, as the key
    has already been advertised in the Certificate message.
    """
    name = "Server RSA_EXPORT parameters"
    fields_desc = [FieldLenField("rsamodlen", None, length_of="rsamod"),
                   StrLenField("rsamod", "",
                               length_from=lambda pkt: pkt.rsamodlen),
                   FieldLenField("rsaexplen", None, length_of="rsaexp"),
                   StrLenField("rsaexp", "",
                               length_from=lambda pkt: pkt.rsaexplen)]

    @crypto_validator
    def fill_missing(self):
        k = PrivKeyRSA()
        k.fill_and_store(modulusLen=512)
        self.tls_session.server_tmp_rsa_key = k
        pubNum = k.pubkey.public_numbers()

        if not self.rsamod:
            self.rsamod = pkcs_i2osp(pubNum.n, k.pubkey.key_size // 8)
        if self.rsamodlen is None:
            self.rsamodlen = len(self.rsamod)

        self.tls_session.kx_group = "rsa%s" % self.rsamodlen

        rsaexplen = math.ceil(math.log(pubNum.e) / math.log(2) / 8.)
        if not self.rsaexp:
            self.rsaexp = pkcs_i2osp(pubNum.e, rsaexplen)
        if self.rsaexplen is None:
            self.rsaexplen = len(self.rsaexp)

    @crypto_validator
    def register_pubkey(self):
        mLen = self.rsamodlen
        m = self.rsamod
        e = self.rsaexp
        self.tls_session.server_tmp_rsa_key = PubKeyRSA((e, m, mLen))
        self.tls_session.kx_group = "rsa%s" % mLen

    def post_dissection(self, pkt):
        try:
            self.register_pubkey()
        except ImportError:
            pass

    def guess_payload_class(self, p):
        return Padding


# Pre-Shared Key

class ServerPSKParams(Packet):
    """
    XXX We provide some parsing abilities for ServerPSKParams, but the
    context operations have not been implemented yet. See RFC 4279.
    Note that we do not cover the (EC)DHE_PSK key exchange,
    which should contain a Server*DHParams after 'psk_identity_hint'.
    """
    name = "Server PSK parameters"
    fields_desc = [FieldLenField("psk_identity_hint_len", None,
                                 length_of="psk_identity_hint", fmt="!H"),
                   StrLenField("psk_identity_hint", "",
                               length_from=lambda pkt: pkt.psk_identity_hint_len)]  # noqa: E501

    def fill_missing(self):
        pass

    def post_dissection(self, pkt):
        pass

    def guess_payload_class(self, p):
        return Padding


###############################################################################
#   Client Key Exchange value                                                 #
###############################################################################

# FFDH/ECDH

class ClientDiffieHellmanPublic(_GenericTLSSessionInheritance):
    """
    If the user provides a value for dh_Yc attribute, we assume he will set
    the pms and ms accordingly and trigger the key derivation on his own.

    XXX As specified in 7.4.7.2. of RFC 4346, we should distinguish the needs
    for implicit or explicit value depending on availability of DH parameters
    in *client* certificate. For now we can only do ephemeral/explicit DH.
    """
    name = "Client DH Public Value"
    fields_desc = [FieldLenField("dh_Yclen", None, length_of="dh_Yc"),
                   StrLenField("dh_Yc", "",
                               length_from=lambda pkt: pkt.dh_Yclen)]

    @crypto_validator
    def fill_missing(self):
        s = self.tls_session
        s.client_kx_privkey = s.client_kx_ffdh_params.generate_private_key()
        pubkey = s.client_kx_privkey.public_key()
        y = pubkey.public_numbers().y
        self.dh_Yc = pkcs_i2osp(y, pubkey.key_size // 8)

        if s.client_kx_privkey and s.server_kx_pubkey:
            pms = s.client_kx_privkey.exchange(s.server_kx_pubkey)
            s.pre_master_secret = pms.lstrip(b"\x00")
            if not s.extms:
                # If extms is set (extended master secret), the key will
                # need the session hash to be computed. This is provided
                # by the TLSClientKeyExchange. Same in all occurrences
                s.compute_ms_and_derive_keys()

    def post_build(self, pkt, pay):
        if not self.dh_Yc:
            try:
                self.fill_missing()
            except ImportError:
                pass
        if self.dh_Yclen is None:
            self.dh_Yclen = len(self.dh_Yc)
        return pkcs_i2osp(self.dh_Yclen, 2) + self.dh_Yc + pay

    def post_dissection(self, m):
        """
        First we update the client DHParams. Then, we try to update the server
        DHParams generated during Server*DHParams building, with the shared
        secret. Finally, we derive the session keys and update the context.
        """
        s = self.tls_session

        # if there are kx params and keys, we assume the crypto library is ok
        if s.client_kx_ffdh_params:
            y = pkcs_os2ip(self.dh_Yc)
            param_numbers = s.client_kx_ffdh_params.parameter_numbers()
            public_numbers = dh.DHPublicNumbers(y, param_numbers)
            s.client_kx_pubkey = public_numbers.public_key(default_backend())

        if s.server_kx_privkey and s.client_kx_pubkey:
            ZZ = s.server_kx_privkey.exchange(s.client_kx_pubkey)
            s.pre_master_secret = ZZ.lstrip(b"\x00")
            if not s.extms:
                s.compute_ms_and_derive_keys()

    def guess_payload_class(self, p):
        return Padding


class ClientECDiffieHellmanPublic(_GenericTLSSessionInheritance):
    """
    Note that the 'len' field is 1 byte longer than with the previous class.
    """
    name = "Client ECDH Public Value"
    fields_desc = [FieldLenField("ecdh_Yclen", None,
                                 length_of="ecdh_Yc", fmt="B"),
                   StrLenField("ecdh_Yc", "",
                               length_from=lambda pkt: pkt.ecdh_Yclen)]

    @crypto_validator
    def fill_missing(self):
        s = self.tls_session
        s.client_kx_privkey = _tls_named_groups_generate(
            s.client_kx_ecdh_params
        )
        # ecdh_Yc follows ECPoint.point format as defined in
        # https://tools.ietf.org/html/rfc8422#section-5.4
        pubkey = s.client_kx_privkey.public_key()
        if isinstance(pubkey, (x25519.X25519PublicKey,
                               x448.X448PublicKey)):
            self.ecdh_Yc = pubkey.public_bytes(
                serialization.Encoding.Raw,
                serialization.PublicFormat.Raw
            )
            if s.client_kx_privkey and s.server_kx_pubkey:
                pms = s.client_kx_privkey.exchange(s.server_kx_pubkey)
        else:
            # uncompressed format of an elliptic curve point
            x = pubkey.public_numbers().x
            y = pubkey.public_numbers().y
            self.ecdh_Yc = (b"\x04" +
                            pkcs_i2osp(x, (pubkey.key_size + 7) // 8) +
                            pkcs_i2osp(y, (pubkey.key_size + 7) // 8))
            if s.client_kx_privkey and s.server_kx_pubkey:
                pms = s.client_kx_privkey.exchange(ec.ECDH(),
                                                   s.server_kx_pubkey)

        if s.client_kx_privkey and s.server_kx_pubkey:
            s.pre_master_secret = pms
            if not s.extms:
                s.compute_ms_and_derive_keys()

    def post_build(self, pkt, pay):
        if not self.ecdh_Yc:
            try:
                self.fill_missing()
            except ImportError:
                pass
        if self.ecdh_Yclen is None:
            self.ecdh_Yclen = len(self.ecdh_Yc)
        return pkcs_i2osp(self.ecdh_Yclen, 1) + self.ecdh_Yc + pay

    def post_dissection(self, m):
        s = self.tls_session

        # if there are kx params and keys, we assume the crypto library is ok
        if s.client_kx_ecdh_params:
            s.client_kx_pubkey = _tls_named_groups_import(
                s.client_kx_ecdh_params,
                self.ecdh_Yc
            )

        if s.server_kx_privkey and s.client_kx_pubkey:
            ZZ = s.server_kx_privkey.exchange(ec.ECDH(), s.client_kx_pubkey)
            s.pre_master_secret = ZZ
            if not s.extms:
                s.compute_ms_and_derive_keys()


# RSA Encryption (standard & export)

class _UnEncryptedPreMasterSecret(Raw):
    """
    When the content of an EncryptedPreMasterSecret could not be deciphered,
    we use this class to represent the encrypted data.
    """
    name = "RSA Encrypted PreMaster Secret (protected)"

    def __init__(self, *args, **kargs):
        kargs.pop('tls_session', None)
        return super(_UnEncryptedPreMasterSecret, self).__init__(*args, **kargs)  # noqa: E501


class EncryptedPreMasterSecret(_GenericTLSSessionInheritance):
    """
    Pay attention to implementation notes in section 7.4.7.1 of RFC 5246.
    """
    name = "RSA Encrypted PreMaster Secret"
    fields_desc = [_TLSClientVersionField("client_version", None,
                                          _tls_version),
                   StrFixedLenField("random", None, 46)]

    @classmethod
    def dispatch_hook(cls, _pkt=None, *args, **kargs):
        if _pkt and 'tls_session' in kargs:
            s = kargs['tls_session']
            if s.server_tmp_rsa_key is None and s.server_rsa_key is None:
                return _UnEncryptedPreMasterSecret
        return EncryptedPreMasterSecret

    def pre_dissect(self, m):
        s = self.tls_session
        tbd = m
        tls_version = s.tls_version
        if tls_version is None:
            tls_version = s.advertised_tls_version
        if tls_version >= 0x0301:
            if len(m) < 2:      # Should not happen
                return m
            tmp_len = struct.unpack("!H", m[:2])[0]
            if len(m) != tmp_len + 2:
                err = "TLS 1.0+, but RSA Encrypted PMS with no explicit length"
                warning(err)
            else:
                tbd = m[2:]
        if s.server_tmp_rsa_key is not None:
            # priority is given to the tmp_key, if there is one
            decrypted = s.server_tmp_rsa_key.decrypt(tbd)
            pms = decrypted[-48:]
        elif s.server_rsa_key is not None:
            decrypted = s.server_rsa_key.decrypt(tbd)
            pms = decrypted[-48:]
        else:
            # the dispatch_hook is supposed to prevent this case
            pms = b"\x00" * 48
            err = "No server RSA key to decrypt Pre Master Secret. Skipping."
            warning(err)

        s.pre_master_secret = pms
        if not s.extms:
            s.compute_ms_and_derive_keys()

        return pms

    def post_build(self, pkt, pay):
        """
        We encrypt the premaster secret (the 48 bytes) with either the server
        certificate or the temporary RSA key provided in a server key exchange
        message. After that step, we add the 2 bytes to provide the length, as
        described in implementation notes at the end of section 7.4.7.1.
        """
        enc = pkt

        s = self.tls_session
        s.pre_master_secret = enc
        if not s.extms:
            s.compute_ms_and_derive_keys()

        if s.server_tmp_rsa_key is not None:
            enc = s.server_tmp_rsa_key.encrypt(pkt, t="pkcs")
        elif s.server_certs is not None and len(s.server_certs) > 0:
            enc = s.server_certs[0].encrypt(pkt, t="pkcs")
        else:
            warning("No material to encrypt Pre Master Secret")

        tmp_len = b""
        tls_version = s.tls_version
        if tls_version is None:
            tls_version = s.advertised_tls_version
        if tls_version >= 0x0301:
            tmp_len = struct.pack("!H", len(enc))
        return tmp_len + enc + pay

    def guess_payload_class(self, p):
        return Padding


# Pre-Shared Key

class ClientPSKIdentity(Packet):
    """
    XXX We provide parsing abilities for ServerPSKParams, but the context
    operations have not been implemented yet. See RFC 4279.
    Note that we do not cover the (EC)DHE_PSK nor the RSA_PSK key exchange,
    which should contain either an EncryptedPMS or a ClientDiffieHellmanPublic.
    """
    name = "Server PSK parameters"
    fields_desc = [FieldLenField("psk_identity_len", None,
                                 length_of="psk_identity", fmt="!H"),
                   StrLenField("psk_identity", "",
                               length_from=lambda pkt: pkt.psk_identity_len)]



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2021 Trend Micro Incorporated

"""
Real-Time Publish-Subscribe Protocol (RTPS) dissection
"""

# scapy.contrib.description = Real-Time Publish-Subscribe Protocol (RTPS)
# scapy.contrib.status = loads
# scapy.contrib.name = rtps

from scapy.contrib.rtps.rtps import *  # noqa F403,F401
from scapy.contrib.rtps.pid_types import * # noqa F403,F401



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information

"""
TLS ciphers.
"""


class CipherError(Exception):
    """
    Raised when .decrypt() or .auth_decrypt() fails.
    """
    pass



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information

# scapy.contrib.description = Postgres PSQL Binary Protocol
# scapy.contrib.status = loads

import struct

from typing import (
    Optional,
    Callable,
    Any,
    Tuple,
)
from scapy.fields import (
    ByteField,
    CharEnumField,
    Field,
    FieldLenField,
    FieldListField,
    IntEnumField,
    PacketListField,
    ShortField,
    SignedIntField,
    SignedShortField,
    StrField,
    StrLenField,
    StrNullField,
)
from scapy.packet import Packet, bind_layers
from scapy.layers.inet import TCP
from scapy.sessions import TCPSession

AUTH_CODES = {
    0: "AuthenticationOk",
    1: "AuthenticationKerberosV4",
    2: "AuthenticationKerberosV5",
    3: "AuthenticationCleartextPassword",
    4: "AuthenticationCryptPassword",
    5: "AuthenticationMD5Password",
    6: "AuthenticationSCMCredential",
    7: "AuthenticationGSS",
    8: "AuthenticationGSSContinue",
    9: "AuthenticationSSPI",
    10: "AuthenticationSASL",
    11: "AuthenticationSASLContinue",
    12: "AuthenticationSASLFinal",
}


class KeepAlive(Packet):
    name = "Keep Alive"
    fields_desc = [
        SignedIntField("len", 4),
    ]


class SSLRequest(Packet):
    name = "SSL request code message"
    fields_desc = [
        FieldLenField("length", None, fmt="I"),
        SignedIntField("request_code", 80877103),
    ]


class _DictStrField(StrField):
    """Takes a dictionary as an argument and packs back into a byte string."""

    def i2m(self, pkt, x):
        if isinstance(x, bytes):
            return x
        if isinstance(x, dict):
            result = bytes()
            for k, v in x.items():
                result += k + b"\x00" + v + b"\x00"
            return result + b"\x00"
        else:
            return super(_DictStrField, self).i2m(pkt, x)

    def i2len(self, pkt, x):
        # type: (Optional[Packet], Any) -> int
        if x is None:
            return 0
        return len(self.i2m(pkt, x))


class Startup(Packet):
    name = "Startup Request Packet"
    fields_desc = [
        FieldLenField(
            "len", None, length_of="options", fmt="I", adjust=lambda pkt, x: x + 8
        ),
        ShortField("protocol_version_major", 3),
        ShortField("protocol_version_minor", 0),
        _DictStrField("options", None),
    ]


class _FieldsLenField(Field[int, int]):
    """Same as FieldLenField but takes a tuple of fields for length_of."""

    __slots__ = ["length_of", "adjust"]

    def __init__(
        self,
        name,  # type: str
        default,  # type: Optional[Any]
        length_of=None,  # type: Optional[Tuple[str]]
        fmt="H",  # type: str
        adjust=lambda pkt, x: x,  # type: Callable[[Packet, int], int]
    ):
        # type: (...) -> None
        super(_FieldsLenField, self).__init__(name, default, fmt)
        self.length_of = length_of
        self.adjust = adjust

    def i2m(self, pkt, x):
        # type: (Optional[Packet], Optional[int]) -> int
        if x is None and pkt is not None:
            if self.length_of is not None:
                f = 0
                for length_of_field in self.length_of:
                    fld, fval = pkt.getfield_and_val(length_of_field)
                    f += fld.i2len(pkt, fval)
            else:
                raise ValueError("Field should have either length_of or count_of")
            x = self.adjust(pkt, f)
        elif x is None:
            x = 0
        return x


def determine_pg_field(pkt, lst, cur, remain):
    key = b""
    if remain:
        key = remain[0:1]  # Python 2/3 compat
    if key in pkt.cls_mapping:
        return pkt.cls_mapping[key]
    elif remain[0:1] == b"\x00" and len(remain) >= 4:
        length = struct.unpack("!I", remain[0:3])[0]
        if length == 0:
            return KeepAlive
        elif length == 8:
            return SSLRequest
        else:
            return Startup
    else:
        return None


class ByteTagField(ByteField):
    def __init__(
        self, default  # type: bytes
    ):
        super(ByteTagField, self).__init__("tag", ord(default))

    def randval(self):
        return ord(self.default)


class _BasePostgres(Packet, TCPSession):
    name = "Regular packet"
    fields_desc = [PacketListField("contents", [], next_cls_cb=determine_pg_field)]

    @classmethod
    def tcp_reassemble(cls, data, metadata):
        if data and data[0:1] == b"\x00":
            length = struct.unpack("!I", data[0:3])[0]
            if length == 8:
                return SSLRequest(data)
            else:
                return Startup(data)
        else:
            return cls(data)


class _ZeroPadding(Packet):
    def extract_padding(self, p):
        return b"", p


class SignedIntStrPair(_ZeroPadding):
    name = "Bytes data"
    fields_desc = [
        FieldLenField("len", 0, fmt="i", length_of="value"),
        StrLenField(
            "data", None, length_from=lambda pkt: pkt.len if pkt.len > 0 else 0
        ),
    ]


class Authentication(_ZeroPadding):
    name = "Authentication Request"
    fields_desc = [
        ByteTagField(b"R"),
        FieldLenField(
            "len", None, length_of="optional", fmt="I", adjust=lambda pkt, x: x + 8
        ),
        IntEnumField("method", default=0, enum=AUTH_CODES),
        StrLenField("optional", None, length_from=lambda pkt: pkt.len - 8),
    ]


class ParameterStatus(_ZeroPadding):
    name = "Parameter Status"
    fields_desc = [
        ByteTagField(b"S"),
        FieldLenField(
            "len",
            None,
            fmt="I",
            length_of=("parameter", "value"),
            adjust=lambda pkt, x: x + 4,
        ),
        StrNullField(
            "parameter",
            "",
        ),
        StrNullField(
            "value",
            "",
        ),
    ]


class Query(_ZeroPadding):
    name = "Simple Query"
    fields_desc = [
        ByteTagField(b"Q"),
        FieldLenField(
            "len", None, length_of="query", fmt="I", adjust=lambda pkt, x: x + 5
        ),
        StrNullField("query", None),
    ]


class CommandComplete(_ZeroPadding):
    name = "Command Completion Response"
    fields_desc = [
        ByteTagField(b"C"),
        FieldLenField(
            "len", None, length_of="cmdtag", fmt="I", adjust=lambda pkt, x: x + 4
        ),
        StrLenField("cmdtag", "", length_from=lambda pkt: pkt.len - 4),
    ]


class BackendKeyData(_ZeroPadding):
    name = "Backend Key Data"
    fields_desc = [
        ByteTagField(b"K"),
        FieldLenField("len", None, fmt="I"),
        SignedIntField("pid", 0),
        SignedIntField("key", 0),
    ]


STATUS_TYPE = {
    b"E": "InFailedTransaction",
    b"I": "Idle",
    b"T": "InTransaction",
}


class ReadyForQuery(_ZeroPadding):
    name = "Ready Signal"
    fields_desc = [
        ByteTagField(b"Z"),
        SignedIntField("len", 6),
        CharEnumField("status", b"I", STATUS_TYPE),
    ]


class ColumnDescription(_ZeroPadding):
    name = "Column Description"
    fields_desc = [
        StrNullField("col", None),
        SignedIntField("tableoid", 0),
        SignedShortField("colno", 0),
        SignedIntField("typeoid", 0),
        SignedShortField("typelen", 0),
        SignedIntField("typemod", 0),
        SignedShortField("format", 0),
    ]


class RowDescription(_ZeroPadding):
    name = "Row Description"
    fields_desc = [
        ByteTagField(b"T"),
        FieldLenField(
            "len", None, fmt="I", length_of="cols", adjust=lambda pkt, x: x + 6
        ),
        SignedShortField("numfields", 0),
        PacketListField(
            "cols",
            [],
            pkt_cls=ColumnDescription,
            count_from=lambda pkt: pkt.numfields,
            length_from=lambda pkt: pkt.len - 6,
        ),
    ]


class DataRow(_ZeroPadding):
    name = "Data Row"
    fields_desc = [
        ByteTagField(b"D"),
        FieldLenField(
            "len", None, fmt="I", length_of="data", adjust=lambda pkt, x: len(pkt) - 1
        ),
        FieldLenField("numfields", 0),
        PacketListField(
            "data",
            [],
            SignedIntStrPair,
            count_from=lambda pkt: pkt.numfields,
        ),
    ]


# See https://www.postgresql.org/docs/current/protocol-error-fields.html
ERROR_FIELD = {
    b"S": "Severity",
    b"V": "SeverityNonLocalized",
    b"C": "Code",
    b"M": "Message",
    b"D": "Detail",
    b"H": "Hint",
    b"P": "Position",
    b"p": "InternalPosition",
    b"q": "InternalQuery",
    b"W": "Where",
    b"s": "SchemaName",
    b"t": "TableName",
    b"c": "ColumnName",
    b"d": "DataTypeName",
    b"n": "ConstraintName",
    b"F": "File",
    b"L": "Line",
    b"R": "Routine",
}


class ErrorResponseField(StrNullField):
    def m2i(self, pkt, x):
        """Unpack into a tuple of Field, Value."""
        i = super(ErrorResponseField, self).m2i(pkt, x)
        i_code = i[0:1]  # Python 2/3 compatible
        return (ERROR_FIELD.get(i_code, i_code), i[1:])


class ErrorResponse(_ZeroPadding):
    name = "Error Response"
    fields_desc = [
        ByteTagField(b"E"),
        FieldLenField(
            "len", None, length_of="error_fields", fmt="I", adjust=lambda pkt, x: x + 5
        ),
        FieldListField(
            "error_fields",
            [],
            ErrorResponseField("value", None),
            length_from=lambda pkt: pkt.len - 5,
        ),
        ByteField("terminator", None),
    ]


class Terminate(_ZeroPadding):
    name = "Termination Request"
    fields_desc = [
        ByteTagField(b"X"),
        SignedIntField("len", 4),
    ]


class _Todo(_ZeroPadding):
    name = "Unsupported message"
    fields_desc = [
        ByteTagField(b"?"),
        FieldLenField("len", None, fmt="I", length_of="body"),
        StrLenField("body", None, length_from=lambda pkt: pkt.len - 4),
    ]


class Bind(_ZeroPadding):
    name = "Bind Request"
    fields_desc = [
        ByteTagField(b"?"),
        FieldLenField(
            "len", None, fmt="I", length_of="body", adjust=lambda pkt, x: len(pkt) - 1
        ),
        StrNullField("destination", ""),
        StrNullField("statement", ""),
        FieldLenField("codes_count", 0, fmt="H", count_of="codes"),
        FieldListField(
            "codes", [], ShortField("", 0), count_from=lambda pkt: pkt.codes_count
        ),
        FieldLenField("values_count", 0, fmt="H", count_of="values"),
        PacketListField(
            "values", [], SignedIntStrPair, count_from=lambda pkt: pkt.values_count
        ),
        FieldLenField("results_count", 0, fmt="H", count_of="results"),
        FieldListField(
            "results", [], ShortField("", 0), count_from=lambda pkt: pkt.results_count
        ),
    ]


class BindComplete(_ZeroPadding):
    name = "Bind Complete"
    fields_desc = [
        ByteTagField(b"2"),
        SignedIntField("len", 4),
    ]


CLOSE_DESCRIBE_TYPE = {b"S": "PreparedStatement", b"P": "Portal"}


class Close(_ZeroPadding):
    name = "Close Request"
    fields_desc = [
        ByteTagField(b"C"),
        FieldLenField(
            "len", None, fmt="I", length_of="statement", adjust=lambda pkt, x: x + 6
        ),
        CharEnumField("close_type", b"S", enum=CLOSE_DESCRIBE_TYPE),
        StrNullField(
            "statement",
            "",
        ),
    ]


class CloseComplete(_ZeroPadding):
    name = "Close Complete"
    fields_desc = [
        ByteTagField(b"3"),
        SignedIntField("len", 4),
    ]


class Describe(_ZeroPadding):
    name = "Describe"
    fields_desc = [
        ByteTagField(b"D"),
        FieldLenField(
            "len", None, fmt="I", length_of="statement", adjust=lambda pkt, x: x + 6
        ),
        CharEnumField("close_type", b"S", enum=CLOSE_DESCRIBE_TYPE),
        StrNullField("statement", ""),
    ]


class EmptyQueryResponse(_ZeroPadding):
    name = "Empty Query Response"
    fields_desc = [
        ByteTagField(b"I"),
        SignedIntField("len", 4),
    ]


class Flush(_ZeroPadding):
    name = "Flush Request"
    fields_desc = [
        ByteTagField(b"H"),
        SignedIntField("len", 4),
    ]


class NoData(_ZeroPadding):
    name = "No Data Response"
    fields_desc = [
        ByteTagField(b"n"),
        SignedIntField("len", 4),
    ]


class ParseComplete(_ZeroPadding):
    name = "Parse Complete Response"
    fields_desc = [
        ByteTagField(b"1"),
        SignedIntField("len", 4),
    ]


class PortalSuspended(_ZeroPadding):
    name = "Portal Suspended Response"
    fields_desc = [
        ByteTagField(b"s"),
        SignedIntField("len", 4),
    ]


class Sync(_ZeroPadding):
    name = "Sync Request"
    fields_desc = [
        ByteTagField(b"S"),
        SignedIntField("len", 4),
    ]


class Parse(_ZeroPadding):
    name = "Parse Request"
    fields_desc = [
        ByteTagField(b"P"),
        FieldLenField("len", None, fmt="I", adjust=lambda pkt, x: len(pkt) - 1),
        StrNullField("destination", ""),
        StrNullField("query", ""),
        FieldLenField("num_param_dtypes", None, fmt="H", count_of="params"),
        FieldListField(
            "params",
            [],
            SignedIntField("param", None),
            count_from=lambda pkt: pkt.num_param_dtypes,
        ),
    ]


class Execute(_ZeroPadding):
    name = "Execute Request"
    fields_desc = [
        ByteTagField(b"E"),
        FieldLenField(
            "len", None, fmt="I", length_of="portal", adjust=lambda pkt, x: x + 9
        ),
        StrNullField(
            "portal",
            "",
        ),
        SignedIntField("rows", 0),
    ]


class PasswordMessage(_ZeroPadding):
    """
    Identifies the message as a password response.
    Note that this is also used for GSSAPI, SSPI and SASL
    response messages. The exact message type can be deduced
    from the context.
    """

    name = "Password Request Response"
    fields_desc = [
        ByteTagField(b"p"),
        FieldLenField(
            "len", None, fmt="I", length_of="password", adjust=lambda pkt, x: x + 4
        ),
        StrLenField("password", None, length_from=lambda pkt: pkt.len - 4),
    ]


class NoticeResponse(_ZeroPadding):
    name = "Notice Response"
    fields_desc = [
        ByteTagField(b"N"),
        FieldLenField(
            "len", None, length_of="notice_fields", fmt="I", adjust=lambda pkt, x: x + 5
        ),
        FieldListField(
            "notice_fields",
            [],
            ErrorResponseField("value", None),
            length_from=lambda pkt: pkt.len - 5,
        ),
        ByteField("terminator", None),
    ]


class NotificationResponse(_ZeroPadding):
    name = "Password Request Response"
    fields_desc = [
        ByteTagField(b"A"),
        _FieldsLenField(
            "len",
            None,
            fmt="I",
            length_of=("channel", "payload"),
            adjust=lambda pkt, x: x + 8,
        ),
        SignedIntField("process_id", 0),
        StrNullField("channel", None),
        StrNullField("payload", None),
    ]


class NegotiateProtocolVersion(_ZeroPadding):
    name = "Negotiate Protocol Version Response"
    fields_desc = [
        ByteTagField(b"v"),
        FieldLenField(
            "len", None, fmt="I", length_of="option", adjust=lambda pkt, x: x + 12
        ),
        SignedIntField("min_minor_version", 0),
        SignedIntField("unrecognized_options", 0),
        StrNullField("option", None),
    ]


class FunctionCallResponse(_ZeroPadding):
    name = "Function Call Response"
    fields_desc = [
        ByteTagField(b"V"),
        FieldLenField(
            "len", None, fmt="I", length_of="result", adjust=lambda pkt, x: x + 8
        ),
        FieldLenField("result_len", None, length_of="result"),
        StrLenField("result", None, length_from=lambda pkt: pkt.result_len),
    ]


class ParameterDescription(_ZeroPadding):
    name = "Parameter Description"
    fields_desc = [
        ByteTagField(b"t"),
        FieldLenField(
            "len", None, fmt="I", length_of="dtypes", adjust=lambda pkt, x: x + 6
        ),
        SignedShortField("dtypes_len", 0),
        FieldListField(
            "dtypes",
            [],
            SignedIntField("dtype", None),
            count_from=lambda pkt: pkt.dtypes_len,
        ),
    ]


class CopyData(_ZeroPadding):
    name = "Copy Data"
    fields_desc = [
        ByteTagField(b"d"),
        FieldLenField(
            "len", None, fmt="I", length_of="data", adjust=lambda pkt, x: x + 4
        ),
        StrLenField("data", None, length_from=lambda pkt: pkt.len - 4),
    ]


class CopyDone(_ZeroPadding):
    name = "Copy Done"
    fields_desc = [
        ByteTagField(b"c"),
        SignedIntField("len", 4),
    ]


class CopyFail(_ZeroPadding):
    name = "Copy Fail Reason"
    fields_desc = [
        ByteTagField(b"f"),
        FieldLenField(
            "len", None, fmt="I", length_of="reason", adjust=lambda pkt, x: x + 4
        ),
        StrLenField("reason", None, length_from=lambda pkt: pkt.len - 4),
    ]


class CancelRequest(Packet):
    name = "Cancel Request"
    fields_desc = [
        SignedIntField("len", 16),
        SignedIntField("request_code", 80877102),
        SignedIntField("process_id", 0),
        SignedIntField("secret", 0),
    ]


class GSSENCRequest(Packet):
    name = "GSSENC Request"
    fields_desc = [
        SignedIntField("len", 8),
        SignedIntField("request_code", 80877104),
    ]


class CopyInResponse(_ZeroPadding):
    name = "Copy in Response"
    fields_desc = [
        ByteTagField(b"G"),
        FieldLenField(
            "len", None, fmt="I", length_of="cols", adjust=lambda pkt, x: x + 7
        ),
        ByteField("format", 0),
        ShortField("ncols", 0),
        FieldListField(
            "cols",
            [],
            ShortField("format", None),
            count_from=lambda pkt: pkt.ncols,
        ),
    ]


class CopyOutResponse(_ZeroPadding):
    name = "Copy out Response"
    fields_desc = [
        ByteTagField(b"H"),
        FieldLenField(
            "len", None, fmt="I", length_of="cols", adjust=lambda pkt, x: x + 7
        ),
        ByteField("format", 0),
        ShortField("ncols", 0),
        FieldListField(
            "cols",
            [],
            ShortField("format", None),
            count_from=lambda pkt: pkt.ncols,
        ),
    ]


class CopyBothResponse(_ZeroPadding):
    name = "Copy both Response"
    fields_desc = [
        ByteTagField(b"W"),
        FieldLenField(
            "len", None, fmt="I", length_of="cols", adjust=lambda pkt, x: x + 7
        ),
        ByteField("format", 0),
        ShortField("ncols", 0),
        FieldListField(
            "cols",
            [],
            ShortField("format", None),
            count_from=lambda pkt: pkt.ncols,
        ),
    ]


FRONTEND_TAG_TO_PACKET_CLS = {
    b"B": Bind,
    b"C": Close,
    b"d": CopyData,
    b"c": CopyDone,
    b"f": CopyFail,
    b"D": Describe,
    b"E": Execute,
    b"H": Flush,
    b"F": _Todo,
    b"P": Parse,
    b"p": PasswordMessage,
    b"Q": Query,
    b"S": Sync,
    b"X": Terminate,
}

BACKEND_TAG_TO_PACKET_CLS = {
    b"R": Authentication,
    b"K": BackendKeyData,
    b"2": BindComplete,
    b"3": CloseComplete,
    b"C": CommandComplete,
    b"d": CopyData,
    b"c": CopyDone,
    b"G": CopyInResponse,
    b"H": CopyOutResponse,
    b"W": CopyBothResponse,
    b"D": DataRow,
    b"I": EmptyQueryResponse,
    b"E": ErrorResponse,
    b"V": FunctionCallResponse,
    b"v": NegotiateProtocolVersion,
    b"n": NoData,
    b"N": NoticeResponse,
    b"A": NotificationResponse,
    b"t": ParameterDescription,
    b"S": ParameterStatus,
    b"1": ParseComplete,
    b"s": PortalSuspended,
    b"Z": ReadyForQuery,
    b"T": RowDescription,
}


class PostgresFrontend(_BasePostgres):
    cls_mapping = FRONTEND_TAG_TO_PACKET_CLS

    @classmethod
    def tcp_reassemble(cls, data, metadata):
        msgs = PostgresFrontend(data)
        if msgs.contents and "Sync" in msgs.contents[-1]:
            return msgs


class PostgresBackend(_BasePostgres):
    cls_mapping = BACKEND_TAG_TO_PACKET_CLS

    @classmethod
    def tcp_reassemble(cls, data, metadata):
        msgs = PostgresBackend(data)
        if msgs.contents and "ReadyForQuery" in msgs.contents[-1]:
            return msgs


bind_layers(TCP, PostgresFrontend, dport=5432)
bind_layers(TCP, PostgresBackend, sport=5432)
